# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/06/2018"


import os
from silx.gui import qt
import logging
from tomwer.core.process.reconstruction.lamino.tofu import (
    SLICE_STACK_TYPE,
    ROTATION_CENTER_TYPE,
    LAMINO_ANGLE_TYPE,
    PSI_ANGLE_TYPE,
)
from tomwer.core.settings import get_lbsram_path
from tomwer.core.utils.char import DELTA_CHAR
from tomwer.gui.reconstruction.lamino.tofu import settings
from tomwer.gui.utils.illustrations import _IllustrationWidget
from .TofuOptionLoader import _TofuOptionLoader, _getterSetter
from .misc import _AngleWidget, _RegionLE, PadlockButton

_logger = logging.getLogger(__name__)


class OutputTofuWidget(_TofuOptionLoader, qt.QWidget):
    """
    Main widgets for the tofu reconstruction
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self._scan_type = "slice stack"
        self.__scanID = None
        self.setLayout(qt.QGridLayout())

        self._controlWidget = qt.QWidget(parent=self)
        self._controlWidget.setLayout(qt.QVBoxLayout())

        self._volumeAngleGrp = VolumeAnglesWidget(parent=self)
        self._stepSizeAndRange = StepGroup(parent=self._controlWidget)

        self._planeDisplay = _IllustrationWidget(parent=self)
        self._planeDisplay.setMinimumSize(qt.QSize(250, 250))
        self._planeDisplay.setSizePolicy(
            qt.QSizePolicy.Preferred, qt.QSizePolicy.Preferred
        )
        self._controlWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._controlWidget.layout().addWidget(self._volumeAngleGrp)
        self._controlWidget.layout().addWidget(self._stepSizeAndRange)
        self._region = RegionGB(parent=self)
        self._controlWidget.layout().addWidget(self._region)

        self._outputWidget = _OutputPathWidget(parent=self)
        self.layout().addWidget(self._outputWidget, 1, 0, 1, 2)

        self.layout().addWidget(self._controlWidget, 0, 0)
        self.layout().addWidget(self._planeDisplay, 0, 1)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 2, 0)

        self._controlWidget.setFixedWidth(500)

        # signal/slot connect
        self._volumeAngleGrp.sigPlaneChanged.connect(self._updateOutput)

        # expose API
        self.getAngles = self._volumeAngleGrp._angles.get
        self.getOutput = self._outputWidget.getOutput
        self.setOutput = self._outputWidget.setOutput
        self.lockOutput = self._outputWidget.lock
        self.forceLbsram = self._outputWidget.forceLbsram
        self.setForceLbsram = self._outputWidget.setForceLbsram
        self.isOutputFolderLocked = self._outputWidget.isLocked
        self.removeTiff = self._outputWidget.removeIfExist
        self.setRemoveTiff = self._outputWidget.setRemoveIfExist
        self.setVolumeAngleX = self._volumeAngleGrp.setVolumeAngleX
        self.setVolumeAngleY = self._volumeAngleGrp.setVolumeAngleY
        self.setVolumeAngleZ = self._volumeAngleGrp.setVolumeAngleZ
        self.getVolumeAngleX = self._volumeAngleGrp.getVolumeAngleX
        self.getVolumeAngleY = self._volumeAngleGrp.getVolumeAngleY
        self.getVolumeAngleZ = self._volumeAngleGrp.getVolumeAngleZ
        self.setRegionSelType = self._stepSizeAndRange.setRegionSelType
        self.setRegion = self._stepSizeAndRange.setRegion
        self.getRegion = self._stepSizeAndRange.getRegion
        self.dryRun = self._outputWidget.dryRun
        self.setHighLimit = self._stepSizeAndRange.setHighLimit
        self.resetHighLimit = self._stepSizeAndRange.resetHighLimit
        self.getZ = self._stepSizeAndRange.getZ
        self.setZ = self._stepSizeAndRange.setZ

        options = {
            "x-region": _getterSetter(
                getter=self._region.getXRegion, setter=self._region.setXRegionFrmStr
            ),
            "y-region": _getterSetter(
                getter=self._region.getYRegion, setter=self._region.setYRegionFrmStr
            ),
            "volume-angle-x": _getterSetter(
                getter=self.getVolumeAngleX, setter=self.setVolumeAngleX
            ),
            "volume-angle-y": _getterSetter(
                getter=self.getVolumeAngleY, setter=self.setVolumeAngleY
            ),
            "volume-angle-z": _getterSetter(
                getter=self.getVolumeAngleZ, setter=self.setVolumeAngleZ
            ),
            "region": _getterSetter(getter=self.getRegion, setter=self.setRegion),
            "output": _getterSetter(getter=self.getOutput, setter=self.setOutput),
            "dry-run": _getterSetter(getter=self.dryRun, setter=self.clearOutput),
            "z": _getterSetter(getter=self.getZ, setter=self.setZ),
            "rm-tif": _getterSetter(getter=self.removeTiff, setter=self.setRemoveTiff),
        }
        _TofuOptionLoader.__init__(self, options=options)

    def clearOutput(self, *args, **kwargs):
        self._outputWidget.clear()

    def _setImg(self, plane):
        assert plane in ("XY", "XZ", "YZ", "manual")
        self._planeDisplay.setImage(plane + "_lamino")

    def loadFromScan(self, scanID):
        self.__scanID = scanID
        self._updateOutput()

    def _updateOutput(self):
        if self.__scanID is None:
            return
        try:
            if self.isOutputFolderLocked() is False:
                self.setOutput(os.path.join(self.__scanID, self._getNameExtension()))
        except Exception as error:
            _logger.error(error)

    def setPlane(self, plane):
        self._volumeAngleGrp._grpPlane.setPlane(plane=plane)

    def _getNameExtension(self):
        """Return the default name extension according to the scan type"""
        if self._scan_type == SLICE_STACK_TYPE:
            if self._volumeAngleGrp._grpPlane.getPlane() == "XY":
                return "xySlice"
            elif self._volumeAngleGrp._grpPlane.getPlane() == "XZ":
                return "xzSlice"
            elif self._volumeAngleGrp._grpPlane.getPlane() == "YZ":
                return "yzSlice"
            elif self._volumeAngleGrp._grpPlane.getPlane() == "manual":
                return "Slice"
        elif self._scan_type == ROTATION_CENTER_TYPE:
            return "xCenter"
        elif self._scan_type == LAMINO_ANGLE_TYPE:
            return "ctAngle"
        elif self._scan_type in (PSI_ANGLE_TYPE, "psi angle"):
            return "rotC"
        else:
            return "Slice"

    def _setScanType(self, scan_type):
        self._scan_type = scan_type
        illustration = scan_type
        if illustration == "slice stack":
            if self._volumeAngleGrp._grpPlane.getPlane() == "XY":
                illustration = "xy slice"
            elif self._volumeAngleGrp._grpPlane.getPlane() == "XZ":
                illustration = "xz slice"
            elif self._volumeAngleGrp._grpPlane.getPlane() == "YZ":
                illustration = "yz slice"
            elif self._volumeAngleGrp._grpPlane.getPlane() == "manual":
                illustration = "manual slice"
        self._planeDisplay.setImage(illustration)

        self._updateOutput()


class PlaneGroup(qt.QGroupBox):

    sigPlaneChanged = qt.Signal(str)
    """Signal emitted when the plane selection changed"""

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, parent=parent, title="Plane selection")

        self.setLayout(qt.QVBoxLayout())
        self._XYPlaneRB = qt.QRadioButton("XY", parent=self)
        self.layout().addWidget(self._XYPlaneRB)
        self._YZPlaneRB = qt.QRadioButton("YZ", parent=self)
        self.layout().addWidget(self._YZPlaneRB)
        self._XZPlaneRB = qt.QRadioButton("XZ", parent=self)
        self.layout().addWidget(self._XZPlaneRB)
        self._autoPlaneRB = qt.QRadioButton("manual", parent=self)
        self.layout().addWidget(self._autoPlaneRB)

        self._rBtns = (
            self._XYPlaneRB,
            self._XZPlaneRB,
            self._YZPlaneRB,
            self._autoPlaneRB,
        )
        self._XYPlaneRB.setChecked(True)

        for btn in self._rBtns:
            btn.toggled.connect(self.__planeChangedCllbck)

            self.setSizePolicy(qt.QSizePolicy.Preferred, qt.QSizePolicy.Preferred)

    def getPlane(self):
        if self._XYPlaneRB.isChecked():
            return "XY"
        if self._YZPlaneRB.isChecked():
            return "YZ"
        if self._XZPlaneRB.isChecked():
            return "XZ"
        if self._autoPlaneRB.isChecked():
            return "manual"

    def setPlane(self, plane):
        assert plane in ("XY", "XZ", "YZ", "manual")
        for w in self._rBtns:
            w.blockSignals(True)
        if plane == "XY":
            self._XYPlaneRB.setChecked(True)
        elif plane == "XZ":
            self._XZPlaneRB.setChecked(True)
        elif plane == "YZ":
            self._YZPlaneRB.setChecked(True)
        else:
            self._autoPlaneRB.setChecked(True)
        self.sigPlaneChanged.emit(plane)
        for w in self._rBtns:
            w.blockSignals(False)

    def __planeChangedCllbck(self):
        self.sigPlaneChanged.emit(self.getPlane())

    def setManualEdition(self):
        self._autoPlaneRB.setChecked(True)


class VolumeAnglesWidget(qt.QWidget):
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self._grpPlane = PlaneGroup(parent=self)
        self.layout().addWidget(self._grpPlane)
        self._angles = _AnglesWidget(parent=self)
        self.layout().addWidget(self._angles)

        # API exposed
        self.sigPlaneChanged = self._grpPlane.sigPlaneChanged
        self.sigAnglesEdited = self._angles.sigAnglesEdited

        self.sigAnglesEdited.connect(self._grpPlane.setManualEdition)
        self.sigPlaneChanged.connect(self._resetAngleFor)

        self.setAngles = self._angles.set
        self.getAngles = self._angles.get
        self.setVolumeAngleX = self._angles.setVolumeAngleX
        self.setVolumeAngleY = self._angles.setVolumeAngleY
        self.setVolumeAngleZ = self._angles.setVolumeAngleZ
        self.getVolumeAngleX = self._angles.getVolumeAngleX
        self.getVolumeAngleY = self._angles.getVolumeAngleY
        self.getVolumeAngleZ = self._angles.getVolumeAngleZ

    def _resetAngleFor(self, plane):
        assert plane in ("XY", "XZ", "YZ", "manual")
        if plane == "manual":
            return

        if plane == "XY":
            self._angles.set(0, 0, 0)
        if plane == "XZ":
            self._angles.set(90, 0, 0)
        if plane == "YZ":
            self._angles.set(0, 90, 0)


class _AnglesWidget(qt.QWidget):
    sigAnglesEdited = qt.Signal()
    """Signal emitted when an angle is edited"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._spacerTop = qt.QWidget(parent=self)
        self._spacerTop.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self._spacerTop)

        self._volAngleX = _AngleWidget(parent=self, name="volume angle x")
        self.layout().addWidget(self._volAngleX)
        self._volAngleY = _AngleWidget(parent=self, name="volume angle y")
        self.layout().addWidget(self._volAngleY)
        self._volAngleZ = _AngleWidget(parent=self, name="volume angle z")
        self.layout().addWidget(self._volAngleZ)
        self._spacerBot = qt.QWidget(parent=self)
        self._spacerBot.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self._spacerBot)

        self._volAngleX.sigEdited.connect(self._haveBeenEdited)
        self._volAngleY.sigEdited.connect(self._haveBeenEdited)
        self._volAngleZ.sigEdited.connect(self._haveBeenEdited)

        # aliases
        self.setVolumeAngleX = self._volAngleX.setAngle
        self.setVolumeAngleY = self._volAngleY.setAngle
        self.setVolumeAngleZ = self._volAngleZ.setAngle
        self.getVolumeAngleX = self._volAngleX.getAngle
        self.getVolumeAngleY = self._volAngleY.getAngle
        self.getVolumeAngleZ = self._volAngleZ.getAngle

    def _haveBeenEdited(self):
        self.sigAnglesEdited.emit()

    def set(self, xAngle, yAngle, zAngle):
        self.blockSignals(True)
        self.setVolumeAngleX(xAngle)
        self.setVolumeAngleY(yAngle)
        self.setVolumeAngleZ(zAngle)
        self.blockSignals(False)

    def get(self):
        """

        :return: tuple of (anglex, angley, anglez)
        """
        return (
            self._volAngleX.getAngle(),
            self._volAngleY.getAngle(),
            self._volAngleZ.getAngle(),
        )


class StepGroup(qt.QGroupBox):
    sigSelectionChanged = qt.Signal()
    """Signal emitted when the selection changed"""

    _NB_SLICE_NAME = "number of slices"
    _RANGE_NAME = "range"

    _TITLE = "Cut selection"

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, parent=parent, title=StepGroup._TITLE)
        self._xcenter = 0.0
        self.__mediumLimit = settings._N_SLICE_LIMITS_MEDIUM
        self.__highLimit = settings._N_SLICE_LIMITS_HIGH

        self._defaultStyleSheet = ""
        self._mediumLimitStyleSheet = (
            "QLineEdit {" + self._convertToStyleSheetColor(settings._COLOR_MEDIUM) + "}"
        )
        self._highLimitStyleSheet = (
            "QLineEdit {" + self._convertToStyleSheetColor(settings._COLOR_HIGH) + "}"
        )

        self.setLayout(qt.QGridLayout())

        # z
        self._z_label = qt.QLabel("z", parent=self)
        self.layout().addWidget(self._z_label, 0, 0)
        self._zSB = qt.QSpinBox(parent=self)
        self._zSB.setMinimum(-99999)
        self._zSB.setValue(0)
        self.layout().addWidget(self._zSB, 0, 1)

        # step size
        self._stepLabel = qt.QLabel("", parent=self)
        self.layout().addWidget(self._stepLabel, 1, 0)

        self._stepSize = qt.QLineEdit("1.0", parent=self)
        validator = qt.QDoubleValidator(parent=self._stepSize)
        validator.setBottom(0)
        self._stepSize.setValidator(validator)
        self.layout().addWidget(self._stepSize, 1, 1, 1, 6)

        style = qt.QApplication.style()
        icon = style.standardIcon(qt.QStyle.SP_MessageBoxWarning)
        self._warningLabel = qt.QLabel("", parent=self)
        self._warningLabel.setPixmap(icon.pixmap(30, state=qt.QIcon.On))
        self._warningLabel.setVisible(False)
        self._warningLabel.setToolTip(
            "Number of slice to reconstruct seems to high, might fail"
        )

        self.layout().addWidget(self._warningLabel, 1, 7, 1, 1)

        # selection mode
        self.layout().addWidget(qt.QLabel("selection mode:"), 2, 0)
        self._selModeTooltips = {
            self._NB_SLICE_NAME: "Will pick n slices from centered in the center, "
            "spaced of step size",
            self._RANGE_NAME: "Will pick slices from a range, each spaced of "
            "step size",
        }
        self._selectionMode = qt.QComboBox(parent=self)
        self._selectionMode.addItem(self._NB_SLICE_NAME)
        idx = self._selectionMode.findText(self._NB_SLICE_NAME)
        self._selectionMode.setItemData(
            idx, self._selModeTooltips[self._NB_SLICE_NAME], qt.Qt.ToolTipRole
        )
        self._selectionMode.addItem(self._RANGE_NAME)
        idx = self._selectionMode.findText(self._RANGE_NAME)
        self._selectionMode.setItemData(
            idx, self._selModeTooltips[self._RANGE_NAME], qt.Qt.ToolTipRole
        )
        self.layout().addWidget(self._selectionMode, 2, 1)

        self._nCutLE = qt.QLineEdit(str(-settings.SLICE_STACK_STEP_SIZE), self)
        validator = qt.QIntValidator(parent=self._nCutLE)
        validator.setBottom(1)
        self._nCutLE.setValidator(validator)
        self.layout().addWidget(self._nCutLE, 2, 2)

        self._fromLabel = qt.QLabel("from:", self)
        self.layout().addWidget(self._fromLabel, 2, 3)
        self._fromLE = qt.QLineEdit(str(-settings.SLICE_STACK_RANGE_HS), self)
        validator = qt.QDoubleValidator(parent=self._fromLE)
        self._fromLE.setValidator(validator)
        self.layout().addWidget(self._fromLE, 2, 4)
        self._toLabel = qt.QLabel("to:", self)
        self.layout().addWidget(self._toLabel, 2, 5)
        self._toLE = qt.QLineEdit(str(settings.SLICE_STACK_RANGE_HS), self)
        validator = qt.QDoubleValidator(parent=self._toLE)
        self._toLE.setValidator(validator)
        self.layout().addWidget(self._toLE, 2, 6)

        self.setRegionSelType(self._RANGE_NAME)
        self._selectionMode.currentIndexChanged[str].connect(self.setRegionSelType)
        self._setStepSizeType("(pixel)")

        # connect signals / SLOT
        for widget in (self._nCutLE, self._stepSize, self._toLE, self._fromLE):
            widget.textChanged.connect(self._updateLimitsColor)

    @staticmethod
    def _convertToStyleSheetColor(color):
        assert type(color) is tuple
        _color = []
        [_color.append(str(c)) for c in color]
        return "color: rgb(" + ",".join(_color) + ")"

    def _setStepSizeType(self, _type):
        assert type(_type) is str
        self._stepLabel.setText(" ".join(("Step size -", DELTA_CHAR, _type)))

    def setRegionSelType(self, selType):
        """

        :param str selType: should be in ('range', 'number of cut')
        """
        assert selType in (self._RANGE_NAME, self._NB_SLICE_NAME)
        index = self._selectionMode.findText(selType)
        assert index >= 0
        self._selectionMode.blockSignals(True)
        self._selectionMode.setCurrentIndex(index)
        self._selectionMode.setToolTip(self._selModeTooltips[selType])
        for w in (self._fromLE, self._toLE, self._toLabel, self._fromLabel):
            w.setVisible(selType == self._RANGE_NAME)
        self._nCutLE.setVisible(selType == self._NB_SLICE_NAME)
        self._selectionMode.blockSignals(False)
        self.sigSelectionChanged.emit()

    def getSelectionType(self):
        """Return active selected mode ('range' or 'number of cut')"""
        return self._selectionMode.currentText()

    def getStepSize(self):
        """Return step size in fofu ref (meter)"""
        return float(self._stepSize.text())

    def getNCut(self):
        if self.getSelectionType() != self._NB_SLICE_NAME:
            return None
        else:
            return int(self._nCutLE.text())

    def setNCut(self, ncut):
        self._nCutLE.setText(str(ncut))

    def getRegion(self):
        if self.getSelectionType() == self._NB_SLICE_NAME:
            start_from = self._xcenter
            if self._zSB.isVisible() is True:
                start_from = self.getZ() or 0.0
            nbCut = self.getNCut()
            _from = -nbCut / 2 * self.getStepSize() + start_from
            _to = nbCut / 2 * self.getStepSize() + start_from
            return _from, _to, self.getStepSize()
        else:
            start_from = 0.0
            if self._zSB.isVisible() is True:
                start_from = self.getZ() or 0.0
            # question: are from, to always integers ?
            return (
                float(self._fromLE.text()) + start_from,
                float(self._toLE.text()) + start_from,
                float(self.getStepSize()),
            )

    def setRegion(self, region):
        if type(region) is str:
            try:
                _from, _to, step_size = region.split(",")
            except Exception:
                _logger.warning(
                    "Fail to setRegion range. given string "
                    "does not fir the standard (from, to, stepSize)"
                )
                return
        else:
            assert type(region) is tuple
            _from, _to, step_size = region
            _from, _to, step_size = str(_from), str(_to), str(step_size)

        self._fromLE.setText(_from)
        self._toLE.setText(_to)
        self._stepSize.setText(step_size)

    def _updateLimitsColor(self, *argv, **kwargs):
        warning = False
        if self._isUpperHightLimit() is True:
            styleSheet = self._highLimitStyleSheet
            warning = True
        elif self._isUpperMediumLimit() is True:
            styleSheet = self._mediumLimitStyleSheet
            warning = True
        else:
            styleSheet = self._defaultStyleSheet

        for widget in (self._fromLE, self._toLE, self._nCutLE, self._stepSize):
            widget.blockSignals(True)

        self.setStyleSheet(styleSheet)

        for widget in (self._fromLE, self._toLE, self._nCutLE, self._stepSize):
            widget.blockSignals(False)

        self._warningLabel.setVisible(warning)

    def _isUpperMediumLimit(self):
        return self._getNSlices() > self.__mediumLimit

    def _isUpperHightLimit(self):
        return self._getNSlices() > self.__highLimit

    def _getNSlices(self):
        """

        :return: number of slice to be reconstructed
        :rtype: int
        """
        if self.getSelectionType() == self._NB_SLICE_NAME:
            return self.getNCut()
        else:
            # as this cast is made during edition, several value can fail on the
            # float conversion or division
            try:
                _from = float(self._fromLE.text())
                _to = float(self._toLE.text())
                _step = float(self.getStepSize())
                return int((_to - _from) / _step)
            except Exception:
                return 1.0

    def setHighLimit(self, val):
        if val == self.__highLimit:
            return
        if val is None:
            self.resetHighLimit()
            return
        if val <= 0.0:
            _logger.error("incoherent height limit: %s" % val)
            return

        self.__highLimit = val
        self._updateLimitsColor()

    def resetHighLimit(self):
        self.__highLimit = settings._N_SLICE_LIMITS_HIGH
        self._updateLimitsColor()

    def getZ(self):
        if self._zSB.isVisible():
            return self._zSB.value()
        else:
            return None

    def setZ(self, value):
        if value is None:
            return
        return self._zSB.setValue(int(value))

    def _setXCenter(self, xcenter):
        self._xcenter = xcenter


class RegionGB(qt.QGroupBox):
    def __init__(self, parent):
        super().__init__("ROI", parent=parent)
        self.setLayout(qt.QVBoxLayout())
        self._xRegion = _RegionLE(name="x ROI", parent=self)
        self.layout().addWidget(self._xRegion)
        self._yRegion = _RegionLE(name="y ROI", parent=self)
        self.layout().addWidget(self._yRegion)

        # expose API
        self.getXRegion = self._xRegion.getRegion
        self.getYRegion = self._yRegion.getRegion
        self.setXRegion = self._xRegion.setRegion
        self.setYRegion = self._yRegion.setRegion
        self.setXRegionFrmStr = self._xRegion.setRegionFromStr
        self.setYRegionFrmStr = self._yRegion.setRegionFromStr


class _OutputPathWidget(qt.QGroupBox):
    class _OptionsWidget(qt.QWidget):
        def __init__(self, parent):
            super().__init__(parent)
            self.setLayout(qt.QHBoxLayout())
            self.layout().setContentsMargins(0, 0, 0, 0)

            self.forceLbsram = qt.QCheckBox("force lbsram", parent=self)
            self.forceLbsram.setToolTip("Make sure reconstruction is saved in lbsram")
            self.layout().addWidget(self.forceLbsram)
            self.forceLbsram.setVisible(os.path.exists(get_lbsram_path()))

            self.removeExisting = qt.QCheckBox('rm "*.tif"', parent=self)
            self.removeExisting.setToolTip("Remove existing files if any")
            self.layout().addWidget(self.removeExisting)
            spacer = qt.QWidget(self)
            spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
            self.layout().addWidget(spacer)

    def __init__(self, parent):
        super().__init__("output folder", parent)
        self.setLayout(qt.QVBoxLayout())
        self.layout().setContentsMargins(4, 4, 4, 4)

        self._checkboxWidget = _OutputPathWidget._OptionsWidget(parent=self)
        self.layout().addWidget(self._checkboxWidget)

        self._qteFolderSelected = qt.QLineEdit("", parent=self)
        style = qt.QApplication.style()
        icon = style.standardIcon(qt.QStyle.SP_DialogOpenButton)
        self._qtbSelectFolder = qt.QPushButton(icon, "", parent=self)
        self._qteFolderSelected.setToolTip(
            "If not output folder is given, then" ' this will generate a "dry run"'
        )

        self._lockButton = PadlockButton(parent=self)
        self._lockButton.toggled.connect(self._dealWithLbsramOptAndLocker)

        leWidget = qt.QWidget(parent=self)
        leWidget.setLayout(qt.QHBoxLayout())
        leWidget.layout().setContentsMargins(0, 0, 0, 0)
        leWidget.layout().addWidget(self._qteFolderSelected)
        leWidget.layout().addWidget(self._qtbSelectFolder)
        leWidget.layout().addWidget(self._lockButton)
        self.layout().addWidget(leWidget)

        # connect signals/SLOT
        self._checkboxWidget.forceLbsram.toggled.connect(self._updateLbsramStatus)
        self._qtbSelectFolder.clicked.connect(self._setFolderPathFrmDiag)
        self._qteFolderSelected.editingFinished.connect(self._checkUpdateforLbsram)

        # add some speaking API
        self.clear = self._qteFolderSelected.clear
        self.isLocked = self._lockButton.isLocked
        self.forceLbsram = self._checkboxWidget.forceLbsram.isChecked
        self.setForceLbsram = self._checkboxWidget.forceLbsram.setChecked
        self.removeIfExist = self._checkboxWidget.removeExisting.isChecked
        self.setRemoveIfExist = self._checkboxWidget.removeExisting.setChecked

    def setOutput(self, text):
        if self.forceLbsram() is True and text.startswith(get_lbsram_path()) is False:
            text = os.path.join(get_lbsram_path(), text.lstrip(os.sep))
        self._qteFolderSelected.setText(text)

    def lock(self, lock=True):
        self._lockButton.setChecked(lock)

    def _setFolderPathFrmDiag(self):
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_() or (len(dialog.selectedFiles()) < 1):
            dialog.close()
            return
        self._qteFolderSelected.setText(dialog.selectedFiles()[0])

    def dryRun(self):
        """Should we execute a dry-run and avoid storing reconstruction"""
        return self.getOutput() is None

    def getOutput(self):
        """Return output folder whre the reconstruction should be saved"""
        if self._qteFolderSelected.text() == "":
            return None
        else:
            return self.__getOutputWithLbsramExtension()

    def _dealWithLbsramOptAndLocker(self, toggle):
        self._checkboxWidget.forceLbsram.setDisabled(toggle)
        if toggle:
            self._checkboxWidget.forceLbsram.setChecked(False)

    def _updateLbsramStatus(self, toggle):
        if toggle is True:
            self._qteFolderSelected.setText(self.__getOutputWithLbsramExtension())

    def __getOutputWithLbsramExtension(self):
        _output = self._qteFolderSelected.text()
        if (
            self.forceLbsram() is True
            and _output.startswith(get_lbsram_path()) is False
        ):
            _output = os.path.join(get_lbsram_path(), _output.lstrip(os.sep))
        return _output

    def _checkUpdateforLbsram(self):
        """Check if the line edit value has to be changed if force lbsram is
        active and the line does not start with it"""
        if self.forceLbsram() is True:
            if not self._qteFolderSelected.text().startswith(get_lbsram_path()):
                self._qteFolderSelected.setText(self.__getOutputWithLbsramExtension())



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "14/02/2020"


import logging
from typing import Optional

import tomwer.core.process.reconstruction.nabu.nabuslices
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.gui.reconstruction.nabu import check
from tomwer.core.settings import isOnLbsram

from contextlib import AbstractContextManager
from silx.gui import qt
from tomwer.synctools.stacks.reconstruction.nabu import NabuSliceProcessStack
from tomwer.core.process.reconstruction.nabu.nabuslices import NabuSlices
from tomwer.gui.reconstruction.nabu.slices import NabuWindow
from tomwer.core.scan.scanbase import TomwerScanBase
from ..utils import WidgetLongProcessing
from orangewidget.settings import Setting
from orangewidget import gui
from tomwer.core.scan.scanbase import _TomwerBaseDock
from orangewidget.widget import Input, Output
from ...orange.managedprocess import SuperviseOW
from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.scan.futurescan import FutureTomwerScan
import copy

_logger = logging.getLogger(__name__)


class NabuOW(WidgetLongProcessing, SuperviseOW):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "nabu slice reconstruction"
    id = "orange.widgets.tomwer.reconstruction.NabuOW.NabuOW"
    description = "This widget will call nabu for running a reconstruction "
    icon = "icons/nabu_2d.svg"
    priority = 12
    keywords = ["tomography", "nabu", "reconstruction", "FBP", "filter"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False
    allows_cycle = True

    # TODO: replace Nabu by default nabu recons by the new orange contrib
    static_input = Setting({"data": None, "nabu_params": None})

    _rpSetting = Setting(dict())
    # this one is keep for backward compatibility. Will be removed on 0.10 I guess

    ewokstaskclass = tomwer.core.process.reconstruction.nabu.nabuslices.NabuSlices

    sigScanReady = qt.Signal(TomwerScanBase)
    "Signal emitted when a scan is ended"

    class Inputs:
        reprocess = Input(
            name="change recons params",
            type=_TomwerBaseDock,
            doc="reconpute slice with different parameters",
        )
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=False,
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

        future_out = Output(
            name="future_data",
            type=FutureTomwerScan,
            doc="data with some remote processing",
        )

    class DialogCM(AbstractContextManager):
        """Simple context manager to hide / show button dialogs"""

        def __init__(self, dialogButtonsBox):
            self._dialogButtonsBox = dialogButtonsBox

        def __enter__(self):
            self._dialogButtonsBox.show()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._dialogButtonsBox.hide()

    def __init__(self, parent=None):
        """
        Widget which read the .hdf5 generated by octave and modify it.
        Then run a subprocess to call octave and run ftseries

        :param bool _connect_handler: True if we want to store the modifications
                                      on the setting. Need for unit test since
                                      keep alive qt widgets.
        :param recons_params: reconsparameter to be used by the FTWidget.
                              If None, some will be created.
        :type: :class:`QReconsParams`
        """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._slurmCluster = None
        self.__exec_for_ci = False
        # processing tool
        self._processingStack = NabuSliceProcessStack(self, process_id=self.process_id)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        # main widget
        self._nabuWidget = NabuWindow(parent=self)
        _layout.addWidget(self._nabuWidget)
        # add button to validate when change reconstruction parameters is
        # called
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        _layout.addWidget(self._buttons)

        # set up
        self._buttons.hide()

        # load settings
        nabu_params = self.static_input.get("nabu_params", None)
        if nabu_params is None:
            nabu_params = self._rpSetting
        if nabu_params != dict():
            print("nabu params are", nabu_params)
            try:
                self._nabuWidget.setConfiguration(nabu_params)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        # expose API
        self.getMode = self._nabuWidget.getMode
        self.setMode = self._nabuWidget.setMode

        # connect signal / slot
        self._processingStack.sigComputationStarted.connect(self._startProcessing)
        self._processingStack.sigComputationEnded.connect(self._endProcessing)
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._nabuWidget.sigConfigChanged.connect(self._updateSettingsVals)

    @Inputs.data
    def process(self, scan: TomwerScanBase):
        assert isinstance(scan, (TomwerScanBase, type(None)))
        if scan is None:
            return
        scan_ = copy.copy(scan)
        scan_.clear_latest_reconstructions()

        if isinstance(scan, HDF5TomoScan):
            if not check.check_flat_series(
                scan_, logger=_logger, user_input=isOnLbsram(scan=scan_)
            ):
                self.Outputs.data.send(scan_)
                return
        if isinstance(scan, HDF5TomoScan):
            if not check.check_dark_series(
                scan_, logger=_logger, user_input=isOnLbsram(scan=scan_)
            ):
                self.Outputs.data.send(scan_)
                return

        _logger.info("add {} to the stack".format(str(scan)))
        # update the reconstruction mode if possible
        self._nabuWidget.setScan(scan_)
        self._processingStack.add(scan_, self.getConfiguration())

    @Inputs.reprocess
    def reprocess(self, scan):
        """Recompute nabu with different parameters"""
        # wait for user to tune the reconstruction
        if scan is None:
            return

        if scan.axis_params is None or scan.axis_params.relative_cor_value is None:
            # try to retrieve last computed cor value from nabu process
            r_cor = NabuSlices.retrieve_last_relative_cor(scan)
            if scan.axis_params is None:
                from tomwer.synctools.axis import QAxisRP

                scan.axis_params = QAxisRP()
            scan.axis_params.set_relative_value(float(r_cor) - scan.dim_1 / 2.0)
            try:
                scan.axis_params.set_relative_value(float(r_cor) - scan.dim_1 / 2.0)
            except Exception:
                pass

        self.show()
        with NabuOW.DialogCM(self._buttons):
            if self.__exec_for_ci is True:
                self._ciExec()
            else:
                if self.exec_():
                    # for now The behavior for reprocessing is the sama as for processing
                    if hasattr(scan, "instance"):
                        self.process(scan=scan.instance)
                    else:
                        self.process(scan=scan)

    @Inputs.cluster_in
    def setCluster(self, slurm_cluster: Optional[SlurmClusterConfiguration]):
        assert isinstance(
            slurm_cluster, (type(None), SlurmClusterConfiguration)
        ), f"Expect None of SlurmClusterConfiguration. Not {type(slurm_cluster)}"
        self._slurmCluster = slurm_cluster

    def _endProcessing(self, scan, future_scan):
        WidgetLongProcessing._endProcessing(self, scan)
        if scan is not None:
            self.Outputs.data.send(scan)
            self.sigScanReady.emit(scan)
        if future_scan is not None:
            self.Outputs.future_out.send(future_scan)

    def setDryRun(self, dry_run):
        self._processingStack.setDryRun(dry_run)

    def _ciExec(self):
        self.activateWindow()
        self.raise_()
        self.show()

    def _replaceExec_(self):
        """used for CI, replace the exec_ call ny"""
        self.__exec_for_ci = True

    def _updateSettingsVals(self):
        self.static_input = {
            "data": None,
            "nabu_params": self.getConfiguration(),
        }
        self._rpSetting = self.getConfiguration()

    def getConfiguration(self):
        config = self._nabuWidget.getConfiguration()
        config["cluster_config"] = self._slurmCluster
        return config

    def setConfiguration(self, config):
        # ignore slurm cluster. Defined by the upper widget
        config.pop("cluster_config", None)
        self._nabuWidget.setConfiguration(config=config)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/08/2017"


import unittest
import tempfile
import os
import shutil
from tomwer.core import utils
from tomwer.core.utils.scanutils import MockEDF, MockHDF5
from tomwer.test.utils import UtilsTest
from tomwer.core.utils.normalization import flatFieldCorrection
import numpy


class TestGetClosestEnergy(unittest.TestCase):
    def setUp(self):
        self.topSrcFolder = tempfile.mkdtemp()
        self.dataSetID = "scan_3_"
        self.dataDir = UtilsTest.getEDFDataset(self.dataSetID)
        self.sourceS3 = os.path.join(self.topSrcFolder, self.dataSetID)
        shutil.copytree(src=os.path.join(self.dataDir), dst=self.sourceS3)

        self.sourceT01 = os.path.join(self.topSrcFolder, "test01")
        shutil.copytree(src=UtilsTest.getEDFDataset("test01"), dst=self.sourceT01)
        self.S3XMLFile = os.path.join(self.sourceS3, "scan_3_.xml")
        self.S3Ref0000 = os.path.join(self.sourceS3, "ref0000_0000.edf")
        self.S3Ref0010 = os.path.join(self.sourceS3, "ref0000_0010.edf")

    def tearDown(self):
        shutil.rmtree(self.topSrcFolder)

    def testEnergyFromEDF(self):
        os.remove(self.S3XMLFile)
        self.assertTrue(
            utils.getClosestEnergy(scan=self.sourceS3, refFile=self.S3Ref0000) == 61
        )
        self.assertTrue(
            utils.getClosestEnergy(scan=self.sourceS3, refFile=self.S3Ref0010) == 61
        )

    def testEnergyFromXML(self):
        os.remove(self.S3Ref0000)
        os.remove(self.S3Ref0010)
        self.assertTrue(
            utils.getClosestEnergy(scan=self.sourceS3, refFile=self.S3Ref0000) == 10
        )
        self.assertTrue(
            utils.getClosestEnergy(scan=self.sourceS3, refFile=self.S3Ref0010) == 10
        )

    def testEnergyFromInfo(self):
        self.assertTrue(utils.getClosestEnergy(scan=self.sourceT01, refFile=None) == 19)

    def testDefaultEnergy(self):
        os.remove(self.S3XMLFile)
        os.remove(self.S3Ref0000)
        os.remove(self.S3Ref0010)
        self.assertTrue(
            utils.getClosestEnergy(scan=self.sourceS3, refFile=self.S3Ref0000) is None
        )
        self.assertTrue(
            utils.getClosestEnergy(scan=self.sourceS3, refFile=self.S3Ref0010) is None
        )


class TestGetClosestSREnergy(unittest.TestCase):
    def setUp(self):
        self.topSrcFolder = tempfile.mkdtemp()
        self.dataSetID = "test10"
        self.dataDir = UtilsTest.getEDFDataset(self.dataSetID)
        self.sourceT10 = os.path.join(self.topSrcFolder, self.dataSetID)
        shutil.copytree(src=os.path.join(self.dataDir), dst=self.sourceT10)
        self.T10XMLFile = os.path.join(self.sourceT10, "test10.xml")
        self.T10InfoFile = os.path.join(self.sourceT10, "test10.info")
        self.T10Ref0000 = os.path.join(self.sourceT10, "ref0000_0000.edf")

    def tearDown(self):
        shutil.rmtree(self.topSrcFolder)

    def testIntenistyFromInfo(self):
        self.assertTrue(
            utils.getClosestSRCurrent(scan_dir=self.sourceT10, refFile=None) == 101.3
        )

    def testDefaultIntensity(self):
        os.remove(self.T10XMLFile)
        os.remove(self.T10InfoFile)
        self.assertTrue(utils.getClosestSRCurrent(scan_dir=self.sourceT10) is None)


class TestMockEDFScan(unittest.TestCase):
    """Test that mock scan are adapted to other unit test"""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.scan_id = os.path.join(self.tmpdir, "myscan")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def testScanEvolution360(self):
        """Test get scan evolution from a mock scan with a 360 range and no
        extra radio"""
        scan = MockEDF.mockScan(
            scanID=self.scan_id,
            nRadio=5,
            nRecons=1,
            nPagRecons=1,
            dim=10,
            scan_range=360,
        )
        scan_dynamic = scan.get_proj_angle_url()
        self.assertEqual(len(scan_dynamic), 5)
        self.assertTrue(0 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[0].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0000.edf"),
        )
        self.assertTrue(180 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[180].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0002.edf"),
        )
        self.assertTrue(360 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[360].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0004.edf"),
        )

    def testScanEvolution180(self):
        """Test get scan evolution from a mock scan with a 180 range and no
        extra radio"""
        scan = MockEDF.mockScan(
            scanID=self.scan_id,
            nRadio=8,
            nRecons=0,
            nPagRecons=0,
            dim=10,
            scan_range=180,
        )
        scan_dynamic = scan.get_proj_angle_url()
        self.assertEqual(len(scan_dynamic), 8)
        self.assertTrue(0 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[0].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0000.edf"),
        )
        self.assertTrue(180 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[180].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0007.edf"),
        )
        self.assertFalse(360 in scan_dynamic)

    def testScanEvolution360Extra4(self):
        """Test get scan evolution from a mock scan with a 360 range and 4
        extra radio"""
        scan = MockEDF.mockScan(
            scanID=self.scan_id,
            nRadio=21,
            nRecons=0,
            nPagRecons=0,
            dim=10,
            scan_range=360,
            n_extra_radio=4,
        )
        scan_dynamic = scan.get_proj_angle_url()
        self.assertEqual(len(scan_dynamic), 21 + 4)
        self.assertTrue(0 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[0].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0000.edf"),
        )
        self.assertTrue(180 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[180].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0010.edf"),
        )
        self.assertTrue(360 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[360].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0020.edf"),
        )
        extra_angles = (0, 90, 180, 270)  # note: 0(1) is the last file acquire
        for iAngle, angle in enumerate(extra_angles):
            angle_id = str(angle) + "(1)"
            self.assertTrue(angle_id in scan_dynamic)
            file_name = os.path.join(
                self.tmpdir, "myscan", "myscan_%04d.edf" % (21 + 4 - 1 - iAngle)
            )
            self.assertEqual(scan_dynamic[angle_id].file_path(), file_name)

    def testScanEvolution180Extra2(self):
        """Test get scan evolution from a mock scan with a 360 range and 3
        extra radios"""
        scan = MockEDF.mockScan(
            scanID=self.scan_id,
            nRadio=4,
            nRecons=2,
            nPagRecons=2,
            dim=10,
            scan_range=180,
            n_extra_radio=2,
        )
        scan_dynamic = scan.get_proj_angle_url()
        self.assertEqual(len(scan_dynamic), 4 + 2)
        self.assertTrue(0 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[0].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0000.edf"),
        )
        self.assertTrue(180 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[180].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0003.edf"),
        )
        self.assertTrue(360 not in scan_dynamic)
        extra_angles = (0, 90)  # note: 0(1) is the last file acquire
        for iAngle, angle in enumerate(extra_angles):
            angle_id = str(angle) + "(1)"
            self.assertTrue(angle_id in scan_dynamic)
            file_name = os.path.join(
                self.tmpdir, "myscan", "myscan_%04d.edf" % (4 + 2 - 1 - iAngle)
            )
            self.assertEqual(scan_dynamic[angle_id].file_path(), file_name)


class TestMockHDF5(unittest.TestCase):
    """Test MockHDF5 to check that the file produced is valid"""

    def setUp(self) -> None:
        self._folder = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self._folder)

    def testSimpleMockCreationOneCall(self):
        """Test mock of an acquisition starting by one dark, then 10 ref,
        then 20 radios, then 10 'final' ref and 2 alignment radio"""
        mock = MockHDF5(
            scan_path=self._folder,
            n_proj=20,
            n_ini_proj=20,
            n_alignement_proj=2,
            create_ini_dark=True,
            create_ini_ref=True,
            create_final_ref=True,
            n_refs=10,
        )
        self.assertTrue(0 in mock.scan.darks.keys())
        self.assertTrue(1 in mock.scan.flats.keys())
        self.assertTrue(10 in mock.scan.flats.keys())


def test_flat_field_normalization():
    """test flatFieldCorrection function"""
    f = numpy.ones((10, 10))
    d = numpy.ones((10, 10))
    imgs = {
        0: numpy.random.random((10, 10)),
        12: numpy.random.random((10, 10)),
    }
    flatFieldCorrection(imgs, d, f)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "25/09/2017"

import re
import numpy

SELECTION_PATTERN = re.compile(r"\d?[:]?\d?")


def selectionIsValid(selection):
    """
    Return true if the given selection as a string is valid

    :return: bool
    """
    assert type(selection) is str
    _selection = selection.replace(" ", "")
    selections = _selection.split(";")
    for sel in selections:
        if not re.match(SELECTION_PATTERN, sel):
            return False
    return True


def getSelection(projections, selection):
    """

    :param str selection:
    :return numpy.ndarray:
    """

    def evalSelection(projections, sel):
        assert type(projections) is numpy.ndarray
        return eval("projections[" + sel + "]")

    assert type(selection) is str
    if selection == "":
        return projections
    _selection = selection.replace(" ", "")
    selections = _selection.split(";")
    if len(selections) == 1:
        return evalSelection(projections, selection)
    else:
        res = None
        for iSel, sel in enumerate(range(len(selections))):
            if iSel == 0:
                res = evalSelection(projections, selections[iSel])
            else:
                res = numpy.append(res, evalSelection(projections, selections[iSel]))
        return res



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["PJ. Gouttenoire", "H. Payno"]
__license__ = "MIT"
__date__ = "23/04/2018"

import logging
import os
import shutil
import tempfile
import unittest
import socket
from glob import glob
from tomwer.core.process.reconstruction.pyhst.pyhstcaller import PyHSTCaller
from tomwer.core.utils.pyhstutils import _findPyHSTVersions, _getPyHSTDir
from tomwer.test.utils import UtilsTest, rebaseParFile
from tomwer.core.utils import getParametersFromParOrInfo
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
from tomwer.core.process.reconstruction.ftseries.params.paganin import PaganinMode
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.utils.device import CudaDevice
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.process.reconstruction.ftseries.params.ft import FixedSliceMode
from tomwer.core import settings
import pytest


logging.disable(logging.INFO)

pyhstVersion = _findPyHSTVersions(_getPyHSTDir())


@pytest.mark.skipif(len(pyhstVersion) == 0, reason="PyHST2 missing")
class TestRecHSTCreation(unittest.TestCase):
    """
    Test that rec HST file are correctly processing
    """

    def setUp(self):
        self.topDir = tempfile.mkdtemp()
        self.dataset = "scan_3_"
        dataTestDir = UtilsTest.getEDFDataset(self.dataset)
        self.targetFolder = os.path.join(self.topDir, self.dataset)
        shutil.copytree(dataTestDir, self.targetFolder)

        self.parFile = os.path.join(self.targetFolder, self.dataset + ".par")
        self.parSliceFile = os.path.join(self.targetFolder, self.dataset + "slice.par")

        for pFile in (self.parFile, self.parSliceFile):
            rebaseParFile(
                filePath=pFile,
                scanOldPath="/lbsram/data/visitor/mi1226/id19/nemoz/henri/",
                scanNewPath=os.path.join(self.topDir),
            )
            rebaseParFile(
                filePath=pFile,
                scanOldPath="/data/visitor/mi1226/id19/nemoz/henri/",
                scanNewPath=os.path.join(self.topDir),
            )

    def tearDown(self):
        for folder in (self.targetFolder, self.topDir):
            shutil.rmtree(folder)

    def testRecFileCreation(self):
        """Check if the PyHST caller is able to create the .rec file"""
        rec_file = os.path.join(self.targetFolder, self.dataset + ".rec")
        par_file = os.path.join(self.targetFolder, self.dataset + ".par")
        PyHSTCaller.make_rec_file(
            rec_file=rec_file, par_file=par_file, recons_params=ReconsParams()
        )
        self.assertTrue(os.path.isfile(rec_file))


class TestParFile(unittest.TestCase):
    """Test the behavior of the `makeParFile` function"""

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def setUp(self):
        # prepare dataset
        self.datasetID = "test10"
        self.tmpdir = tempfile.mkdtemp()
        self.test10_folder = os.path.join(self.tmpdir, self.datasetID)
        test10_path = UtilsTest().getEDFDataset(self.datasetID)
        shutil.copytree(src=test10_path, dst=self.test10_folder)

        # create scan
        self.scan = EDFTomoScan(scan=self.test10_folder)
        self.scan.ftseries_recons_params = ReconsParams(empty=False)
        self.reconsparams = self.scan.ftseries_recons_params
        self.reconsparams.ft.do_test_slice = False
        self.file_prefix = os.path.join(
            self.scan.path, os.path.basename(self.scan.path)
        )
        self.par_file = self.file_prefix + ".par"
        if os.path.exists(self.par_file):
            os.remove(self.par_file)
        self.assertFalse(os.path.exists(self.par_file))

    def testGenericReconsParameters(self):
        pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "dry_run": True,
                "pyhst2_params": self.reconsparams,
            }
        )
        pyhst_caller.run()
        self.assertTrue(os.path.exists(self.par_file))
        par_file_parameters = getParametersFromParOrInfo(self.par_file)
        assert len(par_file_parameters) > 0
        waited_values = {
            "MULTIFRAME": "0",
            "FILE_PREFIX": self.file_prefix,
            "NUM_FIRST_IMAGE": "0",
            "NUM_LAST_IMAGE": "19",
            "NUMBER_LENGTH_VARIES": "NO",
            "LENGTH_OF_NUMERICAL_PART": "4",
            "FILE_POSTFIX": ".edf",
            "FILE_INTERVAL": "1",
            "NUM_IMAGE_1": "2048",
            "NUM_IMAGE_2": "2048",
            "IMAGE_PIXEL_SIZE_1": "3.020000",
            "IMAGE_PIXEL_SIZE_2": "3.020000",
            "SUBTRACT_BACKGROUND": "YES",
            "BACKGROUND_FILE": os.path.join(self.scan.path, "dark.edf"),
            "CORRECT_FLATFIELD": "YES",
            "FLATFIELD_CHANGING": "YES",
            "FLATFIELD_FILE": "refHST0000.edf",
            "FF_PREFIX": os.path.join(self.scan.path, "refHST"),
            "FF_NUM_FIRST_IMAGE": "0",
            "FF_NUM_LAST_IMAGE": "20",
            "FF_NUMBER_LENGTH_VARIES": "NO",
            "FF_LENGTH_OF_NUMERICAL_PART": "4",
            "FF_POSTFIX": ".edf",
            "FF_FILE_INTERVAL": "20",
            "TAKE_LOGARITHM": "YES",
            "ANGLE_BETWEEN_PROJECTIONS": "9.000000",
            "ROTATION_VERTICAL": "YES",
            "ROTATION_AXIS_POSITION": "1024.000000",
            "OUTPUT_SINOGRAMS": "NO",
            "OUTPUT_RECONSTRUCTION": "YES",
            "START_VOXEL_1": "1",
            "START_VOXEL_2": "1",
            "START_VOXEL_3": "1",
            "END_VOXEL_1": "2048",
            "END_VOXEL_2": "2048",
            "END_VOXEL_3": "2048",
            "OVERSAMPLING_FACTOR": "4",
            "ANGLE_OFFSET": "0.000000",
        }
        for key, value in waited_values.items():
            with self.subTest(key=key, value=value):
                self.assertTrue(key.lower() in par_file_parameters)
                self.assertEqual(value, par_file_parameters[key.lower()])

    def testOutputMode(self):
        """Test that the putput mode is correctly write"""
        for file_ext, is_vol in zip((".vol", ".edf"), (True, False)):
            with self.subTest(file_ext=file_ext):
                self.reconsparams.ft.vol_out_file = is_vol
                pyhst_caller = PyHSTCaller(
                    inputs={
                        "data": self.scan,
                        "dry_run": True,
                        "pyhst2_params": self.reconsparams,
                    }
                )
                pyhst_caller.run()
                self.assertTrue(os.path.exists(self.par_file))
                par_file_parameters = getParametersFromParOrInfo(self.par_file)
                self.assertTrue("OUTPUT_FILE".lower() in par_file_parameters)
                output_file = par_file_parameters["OUTPUT_FILE".lower()]
                self.assertEqual(
                    output_file,
                    os.path.join(
                        self.scan.path, os.path.basename(self.scan.path) + file_ext
                    ),
                )

    def testPaganinOff(self):
        """Test par generation if paganin mode is off"""
        self.reconsparams.paganin.mode = PaganinMode.off
        pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "dry_run": True,
                "pyhst2_params": self.reconsparams,
            }
        )
        pyhst_caller.run()
        self.assertTrue(os.path.exists(self.par_file))
        par_file_parameters = getParametersFromParOrInfo(self.par_file)
        self.assertFalse("DO_PAGANIN" in par_file_parameters)
        self.assertFalse("MULTI_PAGANIN_PARS".lower() in par_file_parameters)

    def testPaganinOn(self):
        """Test par generation if paganin mode is on"""
        self.reconsparams.paganin.mode = PaganinMode.on
        self.par_file = os.path.join(self.scan.path, "test10pag_0500.par")
        pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "dry_run": True,
                "pyhst2_params": self.reconsparams,
            }
        )
        pyhst_caller.run()
        self.assertTrue(os.path.exists(self.par_file))

        waited_values = {
            "PAGANIN_Lmicron": "500.000000",
            "PAGANIN_MARGE": "200",
            "DO_OUTPUT_PAGANIN": "0",
            "OUTPUT_PAGANIN_FILE": "paga_cufft",
            "PAGANIN_TRY_CUFFT": "1",
            "PAGANIN_TRY_FFTW": "1",
            "UNSHARP_LOG": "1",
            "PUS": "0.800000",
            "PUC": "3.000000",
            "TRYEDFCONSTANTHEADER": "1",
            "OUTPUT_FILE": self.file_prefix + "pag_0500.edf",
        }
        par_file_parameters = getParametersFromParOrInfo(self.par_file)
        self.assertTrue("DO_PAGANIN".lower() in par_file_parameters)
        self.assertEqual(par_file_parameters["DO_PAGANIN".lower()], "1")

        for key, value in waited_values.items():
            with self.subTest(key=key, value=value):
                self.assertTrue(key.lower() in par_file_parameters)
                self.assertEqual(value, par_file_parameters[key.lower()])

        # test with other values
        self.reconsparams.paganin.mode = PaganinMode.multi
        self.reconsparams.paganin.db = 10
        self.reconsparams.paganin.db2 = 1000
        self.reconsparams.paganin.unsharp_sigma = 0.36
        self.reconsparams.paganin.unsharp_coeff = 0.87
        self.reconsparams.paganin.median_r = 1.25
        self.reconsparams.paganin.dilate = 1.69
        self.reconsparams.paganin.mkeep_soft = True

        pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "dry_run": True,
                "pyhst2_params": self.reconsparams,
            }
        )
        pyhst_caller.run()
        waited_values = {
            "PAGANIN_Lmicron": "%.6f" % self.reconsparams.paganin.db,
            "PAGANIN_MARGE": "200",
            "PUS": "%f" % self.reconsparams.paganin.unsharp_sigma,
            "PUC": "%f" % self.reconsparams.paganin.unsharp_coeff,
        }
        self.par_file = os.path.join(
            self.scan.path,
            os.path.basename(self.scan.path) + "multipag_db0010-db1000.par",
        )
        self.assertTrue(os.path.exists(self.par_file))
        par_file_parameters = getParametersFromParOrInfo(self.par_file)
        for key, value in waited_values.items():
            with self.subTest(key=key, value=value):
                self.assertTrue(key.lower() in par_file_parameters)
                self.assertEqual(value, par_file_parameters[key.lower()])
        self.assertTrue("MULTI_PAGANIN_PARS".lower() in par_file_parameters)

    def testPaganinBoth(self):
        """Test par generation if paganin mode is both"""
        db = self.scan.ftseries_recons_params.paganin.db
        self.par_file = os.path.join(
            self.scan.path, os.path.basename(self.scan.path) + "pag_%04d.par" % db
        )
        self.reconsparams.paganin.mode = PaganinMode.both
        pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "dry_run": True,
                "pyhst2_params": self.reconsparams,
            }
        )
        pyhst_caller.run()
        self.assertTrue(os.path.exists(self.par_file))
        par_file_parameters = getParametersFromParOrInfo(self.par_file)
        self.assertTrue("DO_PAGANIN".lower() in par_file_parameters)
        self.assertEqual(par_file_parameters["DO_PAGANIN".lower()], "1")
        self.assertFalse("MULTI_PAGANIN_PARS".lower() in par_file_parameters)

    def testPaganinMulti(self):
        """Test par generation if paganin mode is multi"""
        db = self.scan.ftseries_recons_params.paganin.db
        db2 = self.scan.ftseries_recons_params.paganin.db2
        self.par_file = os.path.join(
            self.scan.path,
            os.path.basename(self.scan.path) + "multipag_db%04d-db%04d.par" % (db, db2),
        )
        self.reconsparams.paganin.mode = PaganinMode.multi
        pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "dry_run": True,
                "pyhst2_params": self.reconsparams,
            }
        )
        pyhst_caller.run()
        self.assertTrue(os.path.exists(self.par_file))
        par_file_parameters = getParametersFromParOrInfo(self.par_file)
        self.assertTrue("DO_PAGANIN".lower() in par_file_parameters)
        self.assertEqual(par_file_parameters["DO_PAGANIN".lower()], "1")
        self.assertTrue("MULTI_PAGANIN_PARS".lower() in par_file_parameters)
        waited_values = {
            'MULTI_PAGANIN_PARS["DILATE"]': str(self.reconsparams.paganin.dilate),
            'MULTI_PAGANIN_PARS["MEDIANR"]': str(self.reconsparams.paganin.median_r),
            'MULTI_PAGANIN_PARS["THRESHOLD"]': "%1.3f"
            % self.reconsparams.paganin.threshold,
            'MULTI_PAGANIN_PARS["BONE_Lmicron"]': "%.6f"
            % self.reconsparams.paganin.db2,
        }
        for key, value in waited_values.items():
            with self.subTest(key=key, value=value):
                self.assertTrue(key.lower() in par_file_parameters)
                self.assertEqual(value, par_file_parameters[key.lower()])

        # test with some other random values
        self.reconsparams.paganin.dilate = 1
        self.reconsparams.paganin.median_r = 3
        self.reconsparams.paganin.threshold = 400.0
        pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "dry_run": True,
                "pyhst2_params": self.reconsparams,
            }
        )
        pyhst_caller.run()
        par_file_parameters = getParametersFromParOrInfo(self.par_file)
        waited_values = {
            'MULTI_PAGANIN_PARS["DILATE"]': str(self.reconsparams.paganin.dilate),
            'MULTI_PAGANIN_PARS["MEDIANR"]': str(self.reconsparams.paganin.median_r),
            'MULTI_PAGANIN_PARS["THRESHOLD"]': "%1.3f"
            % self.reconsparams.paganin.threshold,
        }
        for key, value in waited_values.items():
            with self.subTest(key=key, value=value):
                self.assertTrue(key.lower() in par_file_parameters)
                self.assertEqual(value, par_file_parameters[key.lower()])


class TestMultipleReconsParams(unittest.TestCase):
    """Test the reconstruction parameters that can have several values."""

    def setUp(self) -> None:
        super(TestMultipleReconsParams, self).setUp()
        self.inputdir = tempfile.mkdtemp()
        mock = MockEDF(scan_path=self.inputdir, n_radio=10, n_ini_radio=10)
        mock.end_acquisition()
        info_file = mock.get_info_file()
        assert os.path.exists(info_file)
        self.scan = EDFTomoScan(scan=mock.scan_path)
        self.scan.ftseries_recons_params = ReconsParams()
        self.scan.ftseries_recons_params.ft.do_test_slice = False
        self.pyhst_caller = PyHSTCaller(
            inputs={
                "pyhst2_params": self.scan.ftseries_recons_params,
                "data": self.scan,
                "dry_run": True,
            }
        )
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 0)

    def tearDown(self) -> None:
        super(TestMultipleReconsParams, self).tearDown()
        shutil.rmtree(self.inputdir)

    def test_off(self):
        """Test that .par file will be generated with the 'default' name
        if no paganin requested"""
        self.scan.ftseries_recons_params.paganin.mode = PaganinMode.off
        self.scan.ftseries_recons_params.paganin.db = [100.0, 200.0]
        # remove all .par file
        self.assertEqual(len(glob(self.scan.path + os.sep + "*.par")), 0)
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 1)
        basename = os.path.basename(self.scan.path)
        self.assertTrue(par_files[0] == os.path.join(self.scan.path, basename + ".par"))

    def test_multi(self):
        """
        Make sure pyhscaller will create several .par file if several
        values of delta/beta are requested and paganin mode is `multi`
        """
        self.scan.ftseries_recons_params.paganin.mode = PaganinMode.multi
        self.scan.ftseries_recons_params.paganin.db = [100.0, 200.0]
        self.scan.ftseries_recons_params.paganin.db2 = [100.0]
        # remove all .par file
        self.assertEqual(len(glob(self.scan.path + os.sep + "*.par")), 0)
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 2)
        basename = os.path.basename(self.scan.path)
        for file_name in (
            basename + "multipag_db0200-db0100.par",
            basename + "multipag_db0100-db0100.par",
        ):
            full_file_path = os.path.join(self.scan.path, file_name)
            self.assertTrue(full_file_path in par_files)

    def test_on_several_value(self):
        """Test .par files generated if paga is `on`"""
        self.scan.ftseries_recons_params.paganin.mode = PaganinMode.on
        self.scan.ftseries_recons_params.paganin.db = [100.0, 200.0]
        # remove all .par file
        self.assertEqual(len(glob(self.scan.path + os.sep + "*.par")), 0)
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 2)
        basename = os.path.basename(self.scan.path)
        for file_name in (basename + "pag_0200.par", basename + "pag_0100.par"):
            full_file_path = os.path.join(self.scan.path, file_name)
            self.assertTrue(full_file_path in par_files)

    def test_both(self):
        """Test .par file generated if paga is `both`"""
        self.scan.ftseries_recons_params.paganin.mode = PaganinMode.both
        self.scan.ftseries_recons_params.paganin.db = [100.0, 200.0]
        # remove all .par file
        self.assertEqual(len(glob(self.scan.path + os.sep + "*.par")), 0)
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 2)
        basename = os.path.basename(self.scan.path)
        for file_name in (basename + "pag_0200.par", basename + "pag_0100.par"):
            full_file_path = os.path.join(self.scan.path, file_name)
            self.assertTrue(full_file_path in par_files)


class TestUtils(unittest.TestCase):
    """Test utils functions"""

    def test_get_gpu_command(self):
        device_0 = CudaDevice(name="device1", id_=0)
        device_1 = CudaDevice(name="device2", id_=1)
        expected_res = socket.gethostname() + ",0"
        self.assertEqual(PyHSTCaller._get_gpu_command((device_0,)), expected_res)

        expected_res = socket.gethostname() + ",1,0"
        self.assertEqual(
            PyHSTCaller._get_gpu_command((device_1, device_0)), expected_res
        )


class TestSlices(unittest.TestCase):
    """Test that the slices parameter is well managed"""

    def setUp(self) -> None:
        super(TestSlices, self).setUp()
        self.inputdir = tempfile.mkdtemp()
        mock = MockEDF(scan_path=self.inputdir, n_radio=10, n_ini_radio=10)
        mock.end_acquisition()
        info_file = mock.get_info_file()
        assert os.path.exists(info_file)
        self.scan = EDFTomoScan(scan=mock.scan_path)
        self.scan.ftseries_recons_params = ReconsParams()
        self.pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "dry_run": True,
                "pyhst2_params": self.scan.ftseries_recons_params,
            }
        )
        self.assertEqual(len(glob(self.scan.path + os.sep + "*.par")), 0)

    def tearDown(self) -> None:
        super(TestSlices, self).tearDown()
        shutil.rmtree(self.inputdir)

    def testNoSlices(self):
        """Test that .par file will be generated with the 'default' name
        if no Paganin requested"""
        self.scan.ftseries_recons_params.ft.do_test_slice = False
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 1)
        basename = os.path.basename(self.scan.path)
        self.assertTrue(par_files[0] == os.path.join(self.scan.path, basename + ".par"))

    def testSingleMiddleSliceNoPag(self):
        """Make sure par file name are correct regarding the do_test and
        the fixed_slice parameter"""
        self.scan.ftseries_recons_params.ft.do_test_slice = True
        self.scan.ftseries_recons_params.ft.fixed_slice = FixedSliceMode.middle
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 2)
        basename = os.path.basename(self.scan.path)
        full_rec_par_file = os.path.join(self.scan.path, basename + ".par")
        self.assertTrue(full_rec_par_file in par_files)

        # check the one with slices
        middle_slice_rec_par_file = (
            basename + "slice_" + str(self.scan.dim_2 // 2).zfill(4)
        )
        middle_slice_rec_par_file = os.path.join(
            self.scan.path, middle_slice_rec_par_file + ".par"
        )
        self.assertTrue(middle_slice_rec_par_file in par_files)

    def testSingleRowNSliceNoPag(self):
        """Make sure par file name are correct regarding the do_test and
        the row_n parameter"""
        self.scan.ftseries_recons_params.ft.do_test_slice = True
        self.scan.ftseries_recons_params.ft.fixed_slice = 2
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 2)
        basename = os.path.basename(self.scan.path)
        full_rec_par_file = os.path.join(self.scan.path, basename + ".par")
        self.assertTrue(full_rec_par_file in par_files)

        # check the one with slices
        middle_slice_rec_par_file = basename + "slice_" + str(2).zfill(4)
        middle_slice_rec_par_file = os.path.join(
            self.scan.path, middle_slice_rec_par_file + ".par"
        )
        self.assertTrue(middle_slice_rec_par_file in par_files)

    def testSingleSliceSeveralPag(self):
        """Test that all .par files for paganin and n rows are created
        if requested"""
        self.scan.ftseries_recons_params.ft.do_test_slice = True
        self.scan.ftseries_recons_params.ft.fixed_slice = 4
        self.scan.ftseries_recons_params.paganin.mode = PaganinMode.both
        self.scan.ftseries_recons_params.paganin.db = [100.0, 200.0]
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 4)
        basename = os.path.basename(self.scan.path)
        # check the .par for full reconstruction
        for file_name in (basename + "pag_0200.par", basename + "pag_0100.par"):
            full_file_path = os.path.join(self.scan.path, file_name)
            self.assertTrue(full_file_path in par_files)

        # check the .par for the middle slice
        # check the .par for full reconstruction
        file_slice_pag_1 = (
            basename
            + "slice_pag_0200_"
            + str(self.scan.ftseries_recons_params.ft.fixed_slice).zfill(4)
            + ".par"
        )
        file_slice_pag_2 = (
            basename
            + "slice_pag_0100_"
            + str(self.scan.ftseries_recons_params.ft.fixed_slice).zfill(4)
            + ".par"
        )
        for file_name in (file_slice_pag_1, file_slice_pag_2):
            full_file_path = os.path.join(self.scan.path, file_name)
            self.assertTrue(full_file_path in par_files)

    def testSeveralSliceNoPag(self):
        """Make sure .par file generated for several slice (using n row)
        are correct if no paganin is activated"""
        slices = (2, 56, 89)
        self.scan.ftseries_recons_params.ft.do_test_slice = True
        self.scan.ftseries_recons_params.ft.fixed_slice = slices
        self.scan.ftseries_recons_params.paganin.mode = PaganinMode.off
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 4)
        basename = os.path.basename(self.scan.path)
        full_rec_par_file = os.path.join(self.scan.path, basename + ".par")
        self.assertTrue(full_rec_par_file in par_files)

        # check the one with slices
        for slice in slices:
            slice_rec_par_file = basename + "slice_" + str(slice).zfill(4)
            slice_rec_par_file = os.path.join(
                self.scan.path, slice_rec_par_file + ".par"
            )
            self.assertTrue(slice_rec_par_file in par_files)

    def testSeveralSliceAndSinglePag(self):
        """Make sure .par file generated for several slice (using n row)
        are correct if only a single paganin is requested"""
        slices = (2, 56, 89)
        self.scan.ftseries_recons_params.ft.do_test_slice = True
        self.scan.ftseries_recons_params.ft.fixed_slice = slices
        self.scan.ftseries_recons_params.paganin.mode = PaganinMode.on
        self.scan.ftseries_recons_params.paganin.db = 100.0
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 4)
        basename = os.path.basename(self.scan.path)
        full_rec_par_file = os.path.join(self.scan.path, basename + "pag_0100.par")
        self.assertTrue(full_rec_par_file in par_files)

        # check the one with slices
        for slice in slices:
            slice_rec_par_file = (
                basename
                + "slice_pag_"
                + "%04d_" % self.scan.ftseries_recons_params.paganin.db
                + str(slice).zfill(4)
            )
            slice_rec_par_file = os.path.join(
                self.scan.path, slice_rec_par_file + ".par"
            )
            self.assertTrue(slice_rec_par_file in par_files)

    def testSeveralSliceAndSeveralPag(self):
        """Make sure .par file generated for several slice (using n row)
        are correct if several paganin are requested"""
        slices = (2, 56, 89)
        db_values = [100.0, 200.0]
        self.scan.ftseries_recons_params.ft.do_test_slice = True
        self.scan.ftseries_recons_params.ft.fixed_slice = slices
        self.scan.ftseries_recons_params.paganin.mode = PaganinMode.on
        self.scan.ftseries_recons_params.paganin.db = db_values
        self.pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 8)
        basename = os.path.basename(self.scan.path)
        for db_value in db_values:
            full_rec_par_file = (
                basename + ("pag_%04d" % db_value).replace(".", "_") + ".par"
            )
            full_rec_par_file = os.path.join(self.scan.path, full_rec_par_file)
            self.assertTrue(full_rec_par_file in par_files)

        # check the one with slices
        for db_value in db_values:
            for slice in slices:
                file_name = basename + ("slice_pag_%04d" % db_value)
                slice_rec_par_file = "_".join((file_name, str(slice).zfill(4)))
                slice_rec_par_file = os.path.join(slice_rec_par_file + ".par")
                slice_rec_par_file = os.path.join(self.scan.path, slice_rec_par_file)
                self.assertTrue(slice_rec_par_file in par_files)


class TestLbsramBehavior(unittest.TestCase):
    """Make sure the .par file will be valid in case we are running under
    with lbsram system"""

    def setUp(self) -> None:
        super(TestLbsramBehavior, self).setUp()
        self.__original_lbsram = settings.get_lbsram_path()
        self.__original_dest = settings.get_dest_path()

        self._lbsram_dir = tempfile.mkdtemp()
        self._dest_dir = tempfile.mkdtemp()
        settings._set_lbsram_path(self._lbsram_dir)
        settings._set_dest_path(self._dest_dir)
        mock = MockEDF(scan_path=self._lbsram_dir, n_radio=10, n_ini_radio=10)
        mock.end_acquisition()
        self.scan = EDFTomoScan(scan=mock.scan_path)
        self.scan.ftseries_recons_params = ReconsParams()
        self.assertEqual(len(glob(self.scan.path + os.sep + "*.par")), 0)

    def tearDown(self) -> None:
        settings._set_lbsram_path(self.__original_lbsram)
        settings._set_dest_path(self.__original_dest)
        for folder in (self._lbsram_dir, self._dest_dir):
            shutil.rmtree(folder)
        super(TestLbsramBehavior, self).tearDown()

    def testSliceParFile(self):
        """Make sure the slice file will start with the LBSRAM_ID"""
        self.scan.ftseries_recons_params.ft.do_test_slice = True
        self.scan.ftseries_recons_params.ft.fixed_slice = 5

        pyhst_caller = PyHSTCaller(
            inputs={
                "data": self.scan,
                "pyhst2_params": self.scan.ftseries_recons_params,
            }
        )
        pyhst_caller.run()
        par_files = glob(self.scan.path + os.sep + "*.par")
        self.assertEqual(len(par_files), 2)
        basename = os.path.basename(self.scan.path)
        full_rec_par_file = os.path.join(self.scan.path, basename + ".par")
        self.assertTrue(full_rec_par_file in par_files)
        full_rec_params = getParametersFromParOrInfo(full_rec_par_file)
        # the .par file for the full volume should have path related
        # to the dest dir
        for key in ("FILE_PREFIX", "BACKGROUND_FILE", "FF_PREFIX", "OUTPUT_FILE"):
            param = full_rec_params[key.lower()]
            self.assertTrue(param.startswith(settings.get_dest_path()))

        middle_slice_rec_par_file = (
            basename
            + "slice_"
            + str(self.scan.ftseries_recons_params.ft.fixed_slice).zfill(4)
        )
        middle_slice_rec_par_file = os.path.join(
            self.scan.path, middle_slice_rec_par_file + ".par"
        )
        self.assertTrue(middle_slice_rec_par_file in par_files)
        slice_rec_params = getParametersFromParOrInfo(middle_slice_rec_par_file)
        # the .par file for slice should have path related to lbsram by default
        for key in ("FILE_PREFIX", "BACKGROUND_FILE", "FF_PREFIX", "OUTPUT_FILE"):
            param = slice_rec_params[key.lower()]
            self.assertTrue(param.startswith(settings.get_lbsram_path()))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "30/09/2020"

from orangewidget import widget, gui
from tomwer.gui.debugtools.objectinspector import ObjectInspector
from orangewidget.widget import Input


class ObjectInspectorOW(widget.OWBaseWidget, openclass=True):
    """
    A simple widget to browse a TomwerScanBase object
    """

    name = "tomwer object browser"
    id = "orangecontrib.tomwer.widgets.debugtools.tomwerscanbasebrowser"
    description = "create on the fly dataset"
    icon = "icons/inspector.png"
    priority = 255
    keywords = ["tomography", "file", "tomwer", "dataset", "debug"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    class Inputs:
        data = Input(name="object", type=object)

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent=parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self.inspector = ObjectInspector(parent=self)
        self._layout.addWidget(self.inspector)

    @Inputs.data
    def setObject(self, obj):
        self.inspector.setObject(obj)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"


from orangecontrib.tomwer.widgets.reconstruction.AxisOW import AxisOW
from tomwer.test.utils import skip_gui_test
from tomwer.test.utils import UtilsTest
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.settings import mock_lsbram
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core import utils
from tomwer.core.utils.scanutils import MockEDF
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
import shutil
import tempfile
import pytest
import time


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestOWAxis(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        self._window = AxisOW()
        self.recons_params = self._window.recons_params
        self.scan_path = UtilsTest.getEDFDataset("D2_H2_T2_h_")
        self._window.getAxis().mode = AxisMode.centered
        self._window.show()
        self.qWaitForWindowExposed(self._window)

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()

    def testAxisLock(self):
        """Test behavior when locking the axis position. Could not be included
        in the tomwer/gui because the lock action is only available for the OW
        """
        assert self._window.getAxis().mode in (AxisMode.centered, AxisMode.global_)
        radio_axis = self._window._widget._axisWidget._radioAxis
        main_widget = self._window._widget._axisWidget
        self.assertFalse(main_widget._controlWidget._lockBut.isLocked())
        self.assertTrue(radio_axis._controlWidget._mainWidget.isEnabled())
        self.mouseClick(main_widget._controlWidget._lockBut, qt.Qt.LeftButton)
        self.qapp.processEvents()
        self.assertTrue(main_widget._controlWidget._lockBut.isLocked())
        # when the lock button is activated we should automatically switch to
        # the manual mode
        self.assertTrue(self._window.getAxis().mode is AxisMode.manual)
        self.assertFalse(radio_axis._controlWidget._mainWidget.isEnabled())


class TestWindowAxisComputation(TestCaseQt):
    @staticmethod
    def _long_computation(scan):
        time.sleep(5)
        return -1

    """Test that the axis widget work correctly"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._mainWindow = AxisOW()
        self.recons_params = self._mainWindow.recons_params
        self._window = self._mainWindow._widget
        self.scan_path = ScanFactory.create_scan_object(
            UtilsTest.getEDFDataset("D2_H2_T2_h_")
        )
        self._mainWindow.show()
        self.qWaitForWindowExposed(self._mainWindow)
        self._mainWindow.setMode("manual")

    def tearDown(self):
        self._mainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._mainWindow.close()
        self._mainWindow = None
        self._window = None
        TestCaseQt.tearDown(self)

    def testFailedComputation(self):
        """Test gui if the axis position fails"""
        self.qapp.processEvents()
        self.recons_params.set_position_frm_par_file("not existing", force=True)
        radio_axis = self._window._axisWidget._radioAxis
        radio_axis.setMode("read")
        self.assertEqual(radio_axis.getMode(), AxisMode.read)
        self._mainWindow.process(self.scan_path)
        self.assertEqual(radio_axis.getMode(), AxisMode.read)
        self.assertEqual(self.recons_params.value_ref_tomwer, None)
        self.assertEqual(self.recons_params.mode, AxisMode.read)

    def testComputationSucceed(self):
        """Test gui if the axis position is correctly computed"""
        self.recons_params.mode = AxisMode.manual
        radio_axis = self._window._axisWidget._radioAxis
        self.assertEqual(radio_axis.getMode(), AxisMode.manual)
        radio_axis.setXShift(2.345)
        self._mainWindow.process(self.scan_path)
        self.qapp.processEvents()
        self.assertEqual(radio_axis.getMode(), AxisMode.manual)
        self.assertEqual(self.recons_params.value_ref_tomwer, 2.345)
        position_info_widget = self._window._axisWidget._controlWidget._positionInfo
        self.assertEqual(position_info_widget._tomwerPositionLabel.text(), "2.345")


global _computation_res
_computation_res = 0


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestAxisStack(TestCaseQt):
    """Test axis computation of a stack of scan"""

    @staticmethod
    def _test_computation(scan):
        global _computation_res
        _computation_res += 1
        return _computation_res

    def setUp(self):
        # not working due to OW
        TestCaseQt.setUp(self)
        self._scan1 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._scan2 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._scan3 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._mainWindow = AxisOW()
        self.recons_params = self._mainWindow._axis_params
        self._mainWindow._skip_exec(True)

        global _computation_res
        _computation_res = 0

        self._mainWindow.patch_calc_method(
            AxisMode.centered, TestAxisStack._test_computation
        )
        self.recons_params.mode = AxisMode.centered

        self._mainWindow.show()
        self.qWaitForWindowExposed(self._mainWindow)

    def tearDown(self):
        utils.mockLowMemory(False)
        mock_lsbram(False)
        self._mainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._mainWindow.close()
        self._mainWindow = None
        self._recons_params = None
        self.qapp.processEvents()
        for scan in (self._scan1, self._scan2, self._scan3):
            shutil.rmtree(scan.path)
        # not working due to OW
        TestCaseQt.tearDown(self)

    def testLowMemory(self):
        """Make sure the axis computation will be skip if we are in low memory"""
        utils.mockLowMemory(True)
        mock_lsbram(True)
        for scan in (self._scan1, self._scan2, self._scan3):
            self.qapp.processEvents()
            self._mainWindow.new_data_in(scan)

        self.assertEqual(self._scan1.axis_params, None)
        self.assertEqual(self._scan2.axis_params, None)
        self.assertEqual(self._scan3.axis_params, None)

    def testUnlockStack(self):
        """Check that all axis position will be computed properly if we set a
        stack of scan"""
        self._mainWindow.recons_params.set_relative_value(1.0)
        for scan in (self._scan1, self._scan2, self._scan3):
            self._mainWindow.process(scan)

        for i in range(5):
            self.qapp.processEvents()
            time.sleep(0.2)
            self.qapp.processEvents()

        self.assertNotEqual(self._scan1.axis_params, None)
        self.assertNotEqual(self._scan2.axis_params, None)
        self.assertNotEqual(self._scan3.axis_params, None)

    def testLockStack(self):
        """Check that axis position will be simply copy if we are in a lock
        stack"""
        self.recons_params.mode = AxisMode.manual
        position_value = 0.36
        self.recons_params.set_relative_value(position_value)

        for scan in (self._scan1, self._scan2, self._scan3):
            self._mainWindow.process(scan)

        for i in range(5):
            self.qapp.processEvents()
            time.sleep(0.2)
            self.qapp.processEvents()

        self.assertEqual(self._scan1.axis_params.value_ref_tomwer, position_value)
        self.assertEqual(self._scan2.axis_params.value_ref_tomwer, position_value)
        self.assertEqual(self._scan3.axis_params.value_ref_tomwer, position_value)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/03/2019"


from .base import _ReconsParam
from ....utils import _assert_cast_to_boolean, _assert_param_instance
from collections.abc import Iterable
from silx.utils.enum import Enum as _Enum
import enum
import typing
import logging

_logger = logging.getLogger(__name__)


@enum.unique
class VolSelMode(_Enum):
    total = 0
    manual = 1
    graphics = 2


@enum.unique
class FixedSliceMode(_Enum):
    middle = -1
    row_n = -3

    @classmethod
    def from_value(cls, value):
        for val in FixedSliceMode:
            if val.name == value:
                return val
        return super(FixedSliceMode, cls).from_value(value)


class FTRP(_ReconsParam):
    """Reconstruction parameters for ftseries (general parameters)"""

    _UNSPLIT_KEYS = ("FIXEDSLICE",)
    "keys that we don't want to treat in `to_unique_recons_set` function"

    def __init__(self):
        _ReconsParam.__init__(self)
        self.__show_proj = False
        self.__show_slice = True
        self.__fixed_slice = FixedSliceMode.middle
        self.__vol_out_file = False
        self.__half_acq = False
        self.__force_half_acq = False
        self.__angle_offset_value = 0.0
        self.__num_part = 4
        self.__fastomo3_version = "fastomo3 3.2"
        self.__correct_spikes_threshold = 0.04
        self.__activate_database = False
        self.__do_test_slice = True
        self.__no_test = False
        self.__zero_off_mask = True
        self.__volume_selection_mode = VolSelMode.total
        self.__record_volume_selection = False
        self.__ring_correction = False
        self.__ddf_correction = False
        self.__fix_header_size = False
        self.__head_directory_to_remove = "/lbsram"
        self.__axis_correction_file = "correct.txt"
        self.__do_axis_correction = False
        self.__force_reconstruction = False
        self.__dff_sigma = 0.0

        self._managed_params = {
            "SHOWPROJ": self.__class__.show_proj,
            "SHOWSLICE": self.__class__.show_slice,
            "FIXEDSLICE": self.__class__.fixed_slice,
            "VOLOUTFILE": self.__class__.vol_out_file,
            "HALF_ACQ": self.__class__.half_acq,
            "FORCE_HALF_ACQ": self.__class__.force_half_acq,
            "ANGLE_OFFSET_VALUE": self.__class__.angle_offset_value,
            "ANGLE_OFFSET": self.__class__.angle_offset,
            "NUM_PART": self.__class__.num_part,
            "VERSION": self.__class__.fastomo3_version,
            "CORRECT_SPIKES_THRESHOLD": self.__class__.correct_spikes_threshold,
            "DATABASE": self.__class__.activate_database,
            "DO_TEST_SLICE": self.__class__.do_test_slice,
            "NO_CHECK": self.__class__.skip_reconstruction_tests,
            "ZEROOFFMASK": self.__class__.set_mask_outside_to_zero,
            "VOLSELECT": self.__class__.volume_selection_mode,
            "VOLSELECTION_REMEMBER": self.__class__.record_volume_selection,
            "RINGSCORRECTION": self.__class__.ring_correction,
            "DFFCORRECTION": self.__class__.double_ff_correction,
            "DFF_SIGMA": self.__class__.double_ff_sigma,
            "FIXHD": self.__class__.fix_header_size,
            "RM_HEAD_DIR": self.__class__.head_directory_to_rm,
            "AXIS_CORRECTION_FILE": self.__class__.axis_correction_file,
            "DO_AXIS_CORRECTION": self.__class__.do_axis_correction,
        }

    @property
    def show_proj(self):
        """show graphical proj during reconstruction"""
        return self.__show_proj

    @show_proj.setter
    def show_proj(self, show):
        _assert_param_instance(show, (bool, int, float))
        _assert_cast_to_boolean(show)
        if self.__show_proj != bool(show):
            self.__show_proj = bool(show)
            self.changed()

    @property
    def show_slice(self):
        """ "show graphical slice during reconstruction"""
        return self.__show_slice

    @show_slice.setter
    def show_slice(self, show):
        _assert_param_instance(show, (bool, int, float))
        _assert_cast_to_boolean(show)
        if self.__show_slice != bool(show):
            self.__show_slice = bool(show)
            self.changed()

    @property
    def fixed_slice(self):
        """which slice(s) to reconstruct"""
        return self.__fixed_slice

    @fixed_slice.setter
    def fixed_slice(self, fixed):
        _assert_param_instance(fixed, (int, str, FixedSliceMode, Iterable))

        # try to see if the value is on of FixedSliceMode
        if isinstance(fixed, str):
            try:
                fixed = FixedSliceMode.from_value(fixed)
            except Exception:
                pass

        if self.__fixed_slice != fixed:
            self.__fixed_slice = fixed
            self.changed()

    def fixed_slice_as_list(self) -> typing.Union[tuple, int]:
        """Return fixed_slice as a list of int"""
        fixed_slice = self.fixed_slice
        if isinstance(fixed_slice, str):
            try:
                if fixed_slice.count(":") == 2:
                    _from, _to, _step = fixed_slice.split(":")
                    _from, _to, _step = int(_from), int(_to), int(_step)
                    if _from > _to:
                        tmp = _to
                        _to = _from
                        _from = tmp
                    res = []
                    while _from <= _to:
                        res.append(_from)
                        _from += _step
                    return tuple(res)
                else:
                    vals = fixed_slice.replace(" ", "")
                    vals = fixed_slice.replace("(", "")
                    vals = fixed_slice.replace(")", "")
                    vals = vals.replace("_", "")
                    vals = vals.replace(";", ",").split(",")
                    res = []
                    for val in vals:
                        try:
                            res.append(int(val))
                        except Exception:
                            pass
                    if len(res) == 1:
                        return res[0]
                    else:
                        return tuple(res)
            except Exception as e:
                _logger.error(e)
        else:
            return fixed_slice

    @property
    def vol_out_file(self):
        """single .vol instead of edf stack"""
        return self.__vol_out_file

    @vol_out_file.setter
    def vol_out_file(self, single_vol):
        _assert_param_instance(single_vol, (bool, int, float))
        _assert_cast_to_boolean(single_vol)
        if self.__vol_out_file != bool(single_vol):
            self.__vol_out_file = bool(single_vol)
            self.changed()

    @property
    def half_acq(self):
        """use half acquisition reconstruction"""
        return self.__half_acq

    @half_acq.setter
    def half_acq(self, half):
        _assert_param_instance(half, (bool, int, float))
        _assert_cast_to_boolean(half)
        if self.__half_acq != bool(half):
            self.__half_acq = bool(half)
            self.changed()

    @property
    def force_half_acq(self):
        """Force half acquisition even if angle is not 360 (from PyHST 2016c)"""
        return self.__force_half_acq

    @force_half_acq.setter
    def force_half_acq(self, force):
        _assert_param_instance(force, (bool, int, float))
        _assert_cast_to_boolean(force)
        self.__force_half_acq = bool(force)
        self.changed()

    @property
    def angle_offset_value(self):
        """finale image rotation angle in degrees"""
        return self.__angle_offset_value

    @angle_offset_value.setter
    def angle_offset_value(self, value):
        assert isinstance(value, (int, float))
        if self.__angle_offset_value != value:
            self.__angle_offset_value = value
            self.changed()

    @property
    def angle_offset(self):
        return self.__angle_offset_value != 0.0

    @angle_offset.setter
    def angle_offset(self, value):
        # nothing to do since it is based on the angle_offset_value. But
        # requested by the structural design choice.
        pass

    @property
    def num_part(self):
        return self.__num_part

    @num_part.setter
    def num_part(self, value):
        """length of the numerical part in the data filenames (for .edf files)"""
        _assert_param_instance(value, (int, float))
        if self.__num_part != int(value):
            self.__num_part = int(value)
            self.changed()

    @property
    def fastomo3_version(self):
        # TODO: this should be removed soon
        return self.__fastomo3_version

    @fastomo3_version.setter
    def fastomo3_version(self, version):
        if self.__fastomo3_version != version:
            self.__fastomo3_version = version
            self.changed()

    @property
    def correct_spikes_threshold(self):
        """threshold above which we have spike"""
        return self.__correct_spikes_threshold

    @correct_spikes_threshold.setter
    def correct_spikes_threshold(self, correct):
        _assert_param_instance(correct, (bool, int, float, str))
        if self.__correct_spikes_threshold != correct:
            self.__correct_spikes_threshold = correct
            self.changed()

    @property
    def activate_database(self):
        """put scan in tomoDB"""
        return self.__activate_database

    @activate_database.setter
    def activate_database(self, activate):
        _assert_param_instance(activate, (bool, int, float))
        _assert_cast_to_boolean(activate)
        if self.__activate_database != bool(activate):
            self.__activate_database = bool(activate)
            self.changed()

    @property
    def do_test_slice(self):
        """reconstruct one test slice"""
        return self.__do_test_slice

    @do_test_slice.setter
    def do_test_slice(self, test):
        _assert_param_instance(test, (bool, int, float))
        _assert_cast_to_boolean(test)
        if self.__do_test_slice != bool(test):
            self.__do_test_slice = bool(test)
            self.changed()

    @property
    def skip_reconstruction_tests(self):
        """force or not reconst of slices in ftseries"""
        return self.__no_test

    @skip_reconstruction_tests.setter
    def skip_reconstruction_tests(self, skip):
        _assert_param_instance(skip, (bool, int, float))
        _assert_cast_to_boolean(skip)
        if self.__no_test != bool(skip):
            self.__no_test = bool(skip)
            self.changed()

    @property
    def force_reconstruction(self):
        """force or not reconst of slices in ftseries"""
        return self.__force_reconstruction

    @force_reconstruction.setter
    def force_reconstruction(self, force):
        _assert_param_instance(force, (bool, int, float))
        _assert_cast_to_boolean(force)
        if self.__force_reconstruction != bool(force):
            self.__force_reconstruction = bool(force)
            self.changed()

    @property
    def set_mask_outside_to_zero(self):
        """Sets to zero the region outside the reconstruction mask"""
        return self.__zero_off_mask

    @set_mask_outside_to_zero.setter
    def set_mask_outside_to_zero(self, set_to_zero):
        _assert_param_instance(set_to_zero, (bool, int, float))
        _assert_cast_to_boolean(set_to_zero)
        if self.__zero_off_mask != set_to_zero:
            self.__zero_off_mask = set_to_zero
            self.changed()

    @property
    def volume_selection_mode(self):
        """how to select volume: total, manual or graphic"""
        return self.__volume_selection_mode

    @volume_selection_mode.setter
    def volume_selection_mode(self, mode):
        _mode = mode
        if type(mode) is str and hasattr(VolSelMode, _mode):
            _mode = getattr(VolSelMode, _mode)
        assert _mode in VolSelMode
        if self.__volume_selection_mode != _mode:
            self.__volume_selection_mode = _mode
            self.changed()

    @property
    def record_volume_selection(self):
        return self.__record_volume_selection

    @record_volume_selection.setter
    def record_volume_selection(self, record):
        _assert_param_instance(record, (bool, int, float))
        _assert_cast_to_boolean(record)
        if bool(record) != self.__record_volume_selection:
            self.__record_volume_selection = bool(record)
            self.changed()

    @property
    def ring_correction(self):
        return self.__ring_correction

    @ring_correction.setter
    def ring_correction(self, _apply):
        _assert_param_instance(_apply, (bool, int, float))
        _assert_cast_to_boolean(_apply)
        if self.__ring_correction != bool(_apply):
            self.__ring_correction = bool(_apply)
            self.changed()

    @property
    def double_ff_correction(self):
        return self.__ddf_correction

    @double_ff_correction.setter
    def double_ff_correction(self, _apply):
        _assert_param_instance(_apply, (bool, int, float))
        _assert_cast_to_boolean(_apply)
        if self.__ddf_correction != bool(_apply):
            self.__ddf_correction = bool(_apply)
            self.changed()

    @property
    def double_ff_sigma(self):
        return self.__dff_sigma

    @double_ff_sigma.setter
    def double_ff_sigma(self, sigma):
        if self.__dff_sigma != float(sigma):
            self.__dff_sigma = float(sigma)
            self.changed()

    @property
    def fix_header_size(self):
        """If true, try fixed header size determination"""
        return self.__fix_header_size

    @fix_header_size.setter
    def fix_header_size(self, fix):
        if self.__fix_header_size != fix:
            self.__fix_header_size = fix
            self.changed()

    @property
    def head_directory_to_rm(self):
        # TODO: should probably be removed, needed by octave but should be
        # managed differently later.
        return self.__head_directory_to_remove

    @head_directory_to_rm.setter
    def head_directory_to_rm(self, _dir):
        assert isinstance(_dir, str)
        if self.__head_directory_to_remove != _dir:
            self.__head_directory_to_remove = _dir
            self.changed()

    @property
    def axis_correction_file(self):
        return self.__axis_correction_file

    @axis_correction_file.setter
    def axis_correction_file(self, _file):
        assert isinstance(_file, str)
        if self.__axis_correction_file != _file:
            self.__axis_correction_file = _file
            self.changed()

    @property
    def do_axis_correction(self):
        return self.__do_axis_correction

    @do_axis_correction.setter
    def do_axis_correction(self, do):
        _assert_param_instance(do, (bool, int, float))
        _assert_cast_to_boolean(do)
        if self.__do_axis_correction != do:
            self.__do_axis_correction = do
            self.changed()

    def to_dict(self):
        _dict = {
            "SHOWPROJ": int(self.show_proj),
            "SHOWSLICE": int(self.show_slice),
            "FIXEDSLICE": self.fixed_slice.name.replace("_", " ")
            if isinstance(self.fixed_slice, FixedSliceMode)
            else str(self.fixed_slice),
            "VOLOUTFILE": int(self.vol_out_file),
            "HALF_ACQ": int(self.half_acq),
            "FORCE_HALF_ACQ": int(self.force_half_acq),
            "ANGLE_OFFSET_VALUE": float(self.angle_offset_value),
            "ANGLE_OFFSET": int(self.angle_offset),
            "NUM_PART": self.num_part,
            "VERSION": self.fastomo3_version,
            "CORRECT_SPIKES_THRESHOLD": self.correct_spikes_threshold,
            "DATABASE": int(self.activate_database),
            "DO_TEST_SLICE": int(self.do_test_slice),
            "NO_CHECK": int(self.skip_reconstruction_tests),
            "ZEROOFFMASK": int(self.set_mask_outside_to_zero),
            "VOLSELECT": self.volume_selection_mode.name,
            "VOLSELECTION_REMEMBER": int(self.record_volume_selection),
            "RINGSCORRECTION": int(self.ring_correction),
            "DFFCORRECTION": float(self.double_ff_correction),
            "DFF_SIGMA": int(self.double_ff_sigma),
            "FIXHD": int(self.fix_header_size),
            "RM_HEAD_DIR": self.head_directory_to_rm,
            "AXIS_CORRECTION_FILE": self.axis_correction_file,
            "DO_AXIS_CORRECTION": int(self.do_axis_correction),
        }
        _dict.update(self.unmanaged_params)
        return _dict

    @staticmethod
    def from_dict(_dict):
        recons_param = FTRP()
        recons_param.load_from_dict(_dict)
        return recons_param

    def load_from_dict(self, _dict):
        self._load_unmanaged_params(_dict=_dict)
        self.show_proj = _dict["SHOWPROJ"]
        self.show_slice = _dict["SHOWSLICE"]
        self.fixed_slice = _dict["FIXEDSLICE"]
        self.vol_out_file = _dict["VOLOUTFILE"]
        self.half_acq = _dict["HALF_ACQ"]
        self.force_half_acq = _dict["FORCE_HALF_ACQ"]
        self.angle_offset_value = _dict["ANGLE_OFFSET_VALUE"]
        self.num_part = _dict["NUM_PART"]
        self.correct_spikes_threshold = _dict["CORRECT_SPIKES_THRESHOLD"]
        self.activate_database = _dict["DATABASE"]
        self.do_test_slice = _dict["DO_TEST_SLICE"]
        self.skip_reconstruction_tests = _dict["NO_CHECK"]
        self.set_mask_outside_to_zero = _dict["ZEROOFFMASK"]
        self.volume_selection_mode = _dict["VOLSELECT"]
        self.record_volume_selection = _dict["VOLSELECTION_REMEMBER"]
        self.ring_correction = _dict["RINGSCORRECTION"]
        self.fix_header_size = _dict["FIXHD"]
        self.head_directory_to_rm = _dict["RM_HEAD_DIR"]
        self.axis_correction_file = _dict["AXIS_CORRECTION_FILE"]
        self.do_axis_correction = _dict["DO_AXIS_CORRECTION"]
        if "DFFCORRECTION" in _dict:
            self.double_ff_correction = _dict["DFFCORRECTION"]
        if "DFF_SIGMA" in _dict:
            self.double_ff_sigma = _dict["DFF_SIGMA"]



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "15/12/2021"


import tempfile
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
from silx.io.url import DataUrl
import shutil
import os
from tomwer.core.utils import scanutils
from tomwer.gui.visualization.reconstructionparameters import ReconstructionParameters


class TestReconstructionParameters(TestCaseQt):
    def setUp(self):
        super().setUp()
        self._window = ReconstructionParameters()
        self._output_dir = tempfile.mkdtemp()
        scan_path = os.path.join(self._output_dir, "my_acquisition")
        self.scan = scanutils.MockHDF5(
            scan_path=scan_path, n_proj=10, n_ini_proj=0
        ).scan

    def tearDown(self):
        shutil.rmtree(self._output_dir)
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        super().tearDown()

    def test(self):
        url = DataUrl(
            file_path=self.scan.master_file,
            data_path=self.scan.entry,
        )
        # in real expect a nabu slice reconstruction file...
        self._window.setUrl(url)
        self._window.show()
        self.qWaitForWindowExposed(self._window)



import sys
from tomwer.app.canvas_launcher.mainwindow import OMain as QMain
from tomwer.core.utils.resource import increase_max_number_file


def main(argv=None):
    increase_max_number_file()
    return QMain().run(argv)


if __name__ == "__main__":
    sys.exit(main())



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from silx.gui import qt, icons as silxicons
from tomwer.core.process.reconstruction.ftseries.params.ft import (
    FixedSliceMode,
    VolSelMode,
)
from tomwer.gui.reconstruction.ftserie.h5editor import H5StructEditor
from tomwer.synctools.ftseries import QReconsParams, _QFTRP
from tomwer.gui.utils.inputwidget import SelectionLineEdit
from tomwer.gui.utils.lineselector import QSliceSelectorDialog
from tomwer.core.scan.scanbase import TomwerScanBase
from silx.io.utils import get_data
from silx.gui.dialog.ImageFileDialog import ImageFileDialog

logger = logging.getLogger(__name__)


class FTWidget(H5StructEditor, qt.QWidget):
    """
    Definition of the Main tab to edit the FT parameters

    :param int rpid: id of the reconstruction parameters edited by the widget
    """

    _ROW_N_INDEX = 1

    def __init__(self, reconsparams, parent=None):
        qt.QWidget.__init__(self, parent)
        H5StructEditor.__init__(self, structID="FT")
        self._recons_params = None
        self.__scan = None
        """used to display some radio"""
        self.setReconsParams(reconsparams)

        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self.__buildReconsOneSlice())
        self.layout().addWidget(self.__buildSelectSlice())
        self.layout().addWidget(self.__buildVolFile())
        self.layout().addWidget(self.__buildVolumeSelection())
        self.layout().addWidget(self.__buildPreviousVolumeSelection())
        self._qcbVolumeSelection.currentIndexChanged.connect(
            self._updatePreviousSelection
        )
        self.layout().addWidget(self.__buildSpikeThreshokd())
        self.layout().addWidget(self.__buildRingCorrection())
        self.layout().addWidget(self.__buildDoubleFFCorrection())
        # spacer
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)
        self._makeConnection()

    def setReconsParams(self, recons_params):
        if isinstance(recons_params, QReconsParams):
            _recons_params = recons_params.ft
        elif isinstance(recons_params, _QFTRP):
            _recons_params = recons_params
        else:
            raise ValueError(
                "recons_params should be an instance of QReconsParam or _QFTRP"
            )

        if self._recons_params:
            self._recons_params.sigChanged.disconnect(self._update_params)
        self._recons_params = _recons_params
        self.load(self._recons_params)
        self._recons_params.sigChanged.connect(self._update_params)

    def setScan(self, scan):
        self.__scan = scan

    def _update_params(self):
        """Update all parameter"""
        self.load(self._recons_params)

    def _makeConnection(self):
        self._qcbDoTestSlice.toggled.connect(self._reconsOneSliceChanged)
        self._qleSliceN.textChanged.connect(self._reconsSelectSliceChanged)
        self._qcbSelectSlice.currentIndexChanged.connect(self._reconsSelectSliceChanged)
        self._qcbEDFStack.toggled.connect(self._volFileChanged)
        self._qcbVolumeSelection.currentIndexChanged.connect(self._volSelectChanged)
        self._qcbVolSelRemenber.toggled.connect(self._volSelectRemenberChanged)
        self._grpThreshold.toggled.connect(self._spikethresholdChanged)
        self._qleThresholdSpikesRm.editingFinished.connect(self._spikethresholdChanged)
        self._qcbRingsCorrection.toggled.connect(self._ringsCorrectionChanged)
        self._qcbDFFCorrection.toggled.connect(self._DFFCorrectionChanged)
        self._onRadioButton.pressed.connect(self._onRadioActivated)

    def __buildReconsOneSlice(self):
        self._qcbDoTestSlice = qt.QCheckBox("Reconstruct slice(s)", parent=self)
        self.linkCheckboxWithH5Variable(self._qcbDoTestSlice, "DO_TEST_SLICE")
        return self._qcbDoTestSlice

    def _reconsOneSliceChanged(self, b):
        if self._isLoading is False:
            self._recons_params._set_parameter_value(
                parameter="DO_TEST_SLICE", value=int(b)
            )
            self._sliceSelectionWidget.setEnabled(b)

    def __buildSelectSlice(self):
        self._sliceSelectionWidget = qt.QWidget()
        self._sliceSelectionWidget.setLayout(qt.QHBoxLayout())
        self._sliceSelectionWidget.layout().addWidget(
            qt.QLabel("Select slices to be reconstructed")
        )
        self._qcbSelectSlice = qt.QComboBox(parent=self._sliceSelectionWidget)
        self._qcbSelectSlice.addItem(FixedSliceMode.middle.name)
        self._qcbSelectSlice.insertItem(
            self._ROW_N_INDEX, FixedSliceMode.row_n.name.replace("_", " ")
        )
        self._qcbSelectSlice.currentIndexChanged.connect(self._setEditableSliceIndex)
        self._sliceSelectionWidget.layout().addWidget(self._qcbSelectSlice)

        # LineEdit position value
        self._qleSliceN = SelectionLineEdit(text="0", parent=self._sliceSelectionWidget)
        self._sliceSelectionWidget.layout().addWidget(self._qleSliceN)
        self._qleSliceN.hide()
        self.linkGroupWithH5Variable(
            group=self._qleSliceN,
            h5ParamName="FIXEDSLICE",
            setter=self._setFixedSlice,
            getter=self._getFixedSlice,
        )
        self._sliceSelectionWidget.layout().addWidget(self._qleSliceN)

        # line selector button
        onRadioIcon = silxicons.getQIcon("window-new")
        self._onRadioButton = qt.QPushButton(onRadioIcon, "on radio", parent=self)
        self._onRadioButton.setToolTip("Open a dialog to pick the rows from" "a radio")
        self._sliceSelectionWidget.layout().addWidget(self._onRadioButton)
        self._onRadioButton.hide()

        self._sliceSelectionWidget.setEnabled(self._recons_params.do_test_slice)
        return self._sliceSelectionWidget

    def _reconsSelectSliceChanged(self, *args, **kwargs):
        if self._isLoading is False:
            old = self._recons_params.blockSignals(True)
            value = self._getFixedSlice()
            self._recons_params._set_parameter_value(
                parameter="FIXEDSLICE", value=value
            )
            self._recons_params.blockSignals(old)

    def _setFixedSlice(self, val):
        if val == FixedSliceMode.middle:
            self._qcbSelectSlice.setCurrentIndex(0)
        else:
            try:
                nbslice = str(val)
            except ValueError as e:
                warning = "Cannot set nb slice value, "
                if len(e.args) > 0:
                    warning += e.args[0]
                logger.warning(warning)
            else:
                self._qcbSelectSlice.setCurrentIndex(self._ROW_N_INDEX)
                self._qleSliceN.setText(str(nbslice))

    def _getFixedSlice(self):
        if self._qcbSelectSlice.currentIndex() == self._ROW_N_INDEX:
            return self._qleSliceN.text()
        else:
            return self._qcbSelectSlice.currentText()

    def _setEditableSliceIndex(self):
        is_row_n_active = self._qcbSelectSlice.currentIndex() == self._ROW_N_INDEX
        self._qleSliceN.setVisible(is_row_n_active)
        self._onRadioButton.setVisible(is_row_n_active)

    def __buildVolFile(self):
        group = qt.QGroupBox("Select output mode :", parent=self)
        group.setLayout(qt.QHBoxLayout())
        self._qcbEDFStack = qt.QRadioButton("Stack of edf files", parent=group)
        group.layout().addWidget(self._qcbEDFStack)
        self._qcbEDFStack.setChecked(True)
        self._qcbSingleVolFile = qt.QRadioButton("Single vol file", parent=group)
        group.layout().addWidget(self._qcbSingleVolFile)
        self.linkGroupWithH5Variable(
            group=group,
            h5ParamName="VOLOUTFILE",
            setter=self._setSingleVolOrStack,
            getter=self._getSingleVolOrStack,
        )
        return group

    def _volFileChanged(self):
        if self._isLoading is False:
            value = self._getSingleVolOrStack()
            self._recons_params._set_parameter_value(
                parameter="VOLOUTFILE", value=value
            )

    def _onRadioActivated(self):
        lineSelection = QSliceSelectorDialog(parent=self)
        if self.__scan is None:
            logger.info("No scan defined, request user for defining the image" "url")
            dialog = ImageFileDialog()
            if dialog.exec_():
                data = dialog.selectedImage()
            else:
                return
        else:
            # TODO: specify lines and selection
            scan = self.__scan
            assert isinstance(scan, TomwerScanBase)
            try:
                projections = scan.projections
                index_radio = list(projections.keys())[len(projections) // 2]
                data = get_data(projections[index_radio])
            except Exception as e:
                logger.error("Fail to load radio data:", str(e))
                return

        lineSelection.setData(data)
        if self._recons_params.fixed_slice not in (0, "0"):
            lineSelection.setSelection(self._recons_params.fixed_slice)
        if lineSelection.exec_() == qt.QDialog.Accepted:
            self._recons_params.fixed_slice = lineSelection.getSelection()

    def __buildVolumeSelection(self):
        widget = qt.QWidget(self)
        widget.setLayout(qt.QVBoxLayout())
        widget.layout().addWidget(qt.QLabel("Volume selection :"))
        self._qcbVolumeSelection = qt.QComboBox(self)
        self._qcbVolumeSelection.addItem(VolSelMode.total.name)
        self._qcbVolumeSelection.addItem(VolSelMode.manual.name)
        self._qcbVolumeSelection.addItem(VolSelMode.graphics.name)
        self.linkComboboxWithH5Variable(self._qcbVolumeSelection, "VOLSELECT")
        widget.layout().addWidget(self._qcbVolumeSelection)
        return widget

    def _volSelectChanged(self):
        if self._isLoading is False:
            value = self._qcbVolumeSelection.currentText()
            self._recons_params._set_parameter_value(parameter="VOLSELECT", value=value)

    def __buildPreviousVolumeSelection(self):
        self._qcbVolSelRemenber = qt.QCheckBox(
            "Remember previous volume selection", parent=self
        )
        self.linkCheckboxWithH5Variable(
            self._qcbVolSelRemenber, h5ParamName="VOLSELECTION_REMEMBER"
        )
        return self._qcbVolSelRemenber

    def _volSelectRemenberChanged(self, b):
        if self._isLoading is False:
            value = int(b)
            self._recons_params._set_parameter_value(
                parameter="VOLSELECTION_REMEMBER", value=value
            )

    def __buildSpikeThreshokd(self):
        self._grpThreshold = qt.QGroupBox(title="Correct spikes", parent=self)
        self._grpThreshold.setCheckable(True)
        self._grpThreshold.setChecked(True)

        self._grpThreshold.setLayout(qt.QHBoxLayout())
        self._grpThreshold.layout().addWidget(qt.QLabel("Threshold for spikes removal"))
        self._qleThresholdSpikesRm = qt.QLineEdit("", parent=self._grpThreshold)
        self._grpThreshold.layout().addWidget(self._qleThresholdSpikesRm)
        self.linkGroupWithH5Variable(
            group=self._grpThreshold,
            h5ParamName="CORRECT_SPIKES_THRESHOLD",
            setter=self._setSpikeThreshold,
            getter=self._getSpikeThreshold,
        )
        return self._grpThreshold

    def _spikethresholdChanged(self):
        if self._isLoading is False:
            value = self._getSpikeThreshold()
            self._recons_params._set_parameter_value(
                parameter="CORRECT_SPIKES_THRESHOLD", value=value
            )

    def __buildRingCorrection(self):
        self._qcbRingsCorrection = qt.QCheckBox(
            "Perform rings corrections in PyHST", parent=self
        )
        self.linkCheckboxWithH5Variable(self._qcbRingsCorrection, "RINGSCORRECTION")
        return self._qcbRingsCorrection

    def _ringsCorrectionChanged(self, b):
        if self._isLoading is False:
            value = int(b)
            self._recons_params._set_parameter_value(
                parameter="RINGSCORRECTION", value=value
            )

    def __buildDoubleFFCorrection(self):
        # build the widget
        self.__dff_widget = qt.QWidget(parent=self)
        self.__dff_widget.setLayout(qt.QHBoxLayout())
        self.__dff_widget.setContentsMargins(0, 0, 0, 0)
        self.__dff_widget.layout().setContentsMargins(0, 0, 0, 0)

        self._qcbDFFCorrection = qt.QCheckBox("Perform double flat field correction")
        self.linkCheckboxWithH5Variable(self._qcbDFFCorrection, "DFFCORRECTION")
        self.__dff_widget.layout().addWidget(self._qcbDFFCorrection)

        label = qt.QLabel("sigma:", self)
        label.setAlignment(qt.Qt.AlignRight)
        self.__dff_widget.layout().addWidget(label)
        self._qdsbDFFSigma = qt.QDoubleSpinBox(parent=self)
        self._qdsbDFFSigma.setMinimum(0.0)
        self._qdsbDFFSigma.setDecimals(2)
        self._qdsbDFFSigma.setSingleStep(0.1)
        tooltip = (
            "high-pass filtering on double flatfield with this value " "of 'sigma'"
        )
        self._qdsbDFFSigma.setToolTip(tooltip)
        self.__dff_widget.layout().addWidget(self._qdsbDFFSigma)
        self.linkGroupWithH5Variable(
            self._qcbDFFCorrection,
            "DFF_SIGMA",
            setter=self._setDFFSigma,
            getter=self._getDFFSigma,
        )
        # set up
        self._qdsbDFFSigma.setEnabled(False)
        # connect option and spin box
        self._qcbDFFCorrection.toggled.connect(self._qdsbDFFSigma.setEnabled)
        return self.__dff_widget

    def _setDFFSigma(self, sigma):
        self._qdsbDFFSigma.setValue(sigma)

    def _getDFFSigma(self):
        return self._qdsbDFFSigma.value()

    def _DFFSigmaChanged(self, b):
        if self._isLoading is False:
            value = float(b)
            self._recons_params._set_parameter_value(parameter="DFF_SIGMA", value=value)

    def _DFFCorrectionChanged(self, b):
        if self._isLoading is False:
            value = int(b)
            self._recons_params._set_parameter_value(
                parameter="DFFCORRECTION", value=value
            )

    def _getSingleVolOrStack(self):
        """return:
        - 0 if we want to use a single file for volume
        - 1 if we want to use a stack of edf file for volume
        """
        return int(not self._qcbEDFStack.isChecked())

    def _setSingleVolOrStack(self, val):
        """Set the value of the volume file selection"""
        assert isinstance(val, (int, float, bool))
        assert int(val) in [0, 1]
        if val == 0:
            self._qcbEDFStack.setChecked(True)
        else:
            self._qcbSingleVolFile.setChecked(True)

    def _enableVolumeSelection(self, disable):
        """Do we give the user access to the volume selection or is it set to
        total
        """
        self._qcbVolumeSelection.setEnabled(not disable)
        if not disable:
            self._qcbVolumeSelection.setCurrentIndex(0)

    def _updatePreviousSelection(self):
        self._qcbVolSelRemenber.setEnabled(self._qcbVolumeSelection.currentIndex() == 1)

    def _setSpikeThreshold(self, value):
        if type(value) is str:
            assert value.lower() == "inf"
            self._grpThreshold.setChecked(False)
        else:
            assert type(value) is float
            self._grpThreshold.setChecked(True)
            self._qleThresholdSpikesRm.setText(str(value))

    def _getSpikeThreshold(self):
        if self._grpThreshold.isChecked():
            return float(self._qleThresholdSpikesRm.text())
        else:
            return "Inf"



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "12/04/2021"

import time

from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
from tomwer.core.utils.scanutils import MockHDF5
from orangecontrib.tomwer.widgets.reconstruction.SAAxisOW import SAAxisOW as _SAAxisOW
from tomwer.core.process.reconstruction.scores import ComputedScore
from processview.core.manager import ProcessManager, DatasetState
from silx.io.url import DataUrl
import h5py
import tempfile
import shutil
import logging
import numpy
import uuid
import os
from tomwer.core import utils, settings

logger = logging.getLogger(__name__)


class SAAxisOW(_SAAxisOW):
    def __init__(self, parent=None):
        self._scans_finished = []
        super().__init__(parent)

    def processing_finished(self, scan):
        # TODO: add message processing finished
        self._scans_finished.append(scan)

    @property
    def scans_finished(self):
        return self._scans_finished

    def close(self):
        self._scans_finished = {}
        super().close()


class TestProcessing(TestCaseQt):

    DIM = 100

    def setUp(self):
        super().setUp()
        utils.mockLowMemory(False)
        settings.mock_lsbram(False)
        self._source_dir = tempfile.mkdtemp()

        def create_scan(folder_name):
            _dir = os.path.join(self._source_dir, folder_name)
            return MockHDF5(
                scan_path=_dir,
                n_ini_proj=20,
                n_proj=20,
                n_alignement_proj=2,
                create_final_ref=False,
                create_ini_dark=True,
                create_ini_ref=True,
                n_refs=1,
                dim=self.DIM,
            ).scan

        # create scans
        self.scan_1 = create_scan("scan_1")
        self.scan_2 = create_scan("scan_2")
        self.scan_3 = create_scan("scan_3")
        self._process_manager = ProcessManager()

        self.widget = SAAxisOW()
        self.widget.show()

        def patch_score(*args, **kwargs):
            data = numpy.random.random(TestProcessing.DIM * TestProcessing.DIM)
            data = data.reshape(TestProcessing.DIM, TestProcessing.DIM)
            slice_file_path = os.path.join(
                self._source_dir, str(uuid.uuid1()) + ".hdf5"
            )
            data_url = DataUrl(
                file_path=slice_file_path, data_path="data", scheme="silx"
            )
            with h5py.File(slice_file_path, mode="a") as h5f:
                h5f["data"] = data
            return data_url, ComputedScore(
                tv=numpy.random.random(),
                std=numpy.random.random(),
            )

        self.widget._widget._processing_stack.patch_processing(patch_score)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.stop()
        self.widget.close()
        self.widget = None
        self.qapp.processEvents()
        shutil.rmtree(self._source_dir)
        super().tearDown()

    def testAutoFocusUnlock(self):
        self.widget.lockAutofocus(False)

        def manual_processing():
            self.widget.load_sinogram()
            self.widget.compute()
            self.qapp.processEvents()
            self.widget.wait_processing(5000)
            self.qapp.processEvents()

        self.widget.process(self.scan_1)
        manual_processing()
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_1.get_dataset_identifier(),
                process=self.widget,
            ),
            DatasetState.WAIT_USER_VALIDATION,
        )

        self.widget.process(self.scan_2)
        manual_processing()
        self.assertEqual(len(self.widget.scans_finished), 0)
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_1.get_dataset_identifier(),
                process=self.widget,
            ),
            DatasetState.SKIPPED,
        )
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_2.get_dataset_identifier(),
                process=self.widget,
            ),
            DatasetState.WAIT_USER_VALIDATION,
        )
        self.widget.process(self.scan_3)
        manual_processing()
        self.widget.validateCurrentScan()
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_3.get_dataset_identifier(),
                process=self.widget,
            ),
            DatasetState.SUCCEED,
        )
        # insure a cor has been registered
        self.assertNotEqual(self.scan_3.axis_params.relative_cor_value, None)

    def testTestLbsram(self):
        utils.mockLowMemory(True)
        settings.mock_lsbram(True)
        for scan in (self.scan_1, self.scan_2, self.scan_3):
            self.widget.process(scan)
            self.widget.wait_processing(5000)
            self.qapp.processEvents()

        for scan in (self.scan_1, self.scan_2, self.scan_3):
            with self.subTest(scan=str(scan)):
                self.assertEqual(
                    self._process_manager.get_dataset_state(
                        dataset_id=scan.get_dataset_identifier(),
                        process=self.widget,
                    ),
                    DatasetState.SKIPPED,
                )

    def testAutoFocusLock(self):
        self.widget.lockAutofocus(True)
        for scan in (self.scan_1, self.scan_2, self.scan_3):
            self.widget.process(scan)
            self.widget.wait_processing(10000)
            self.qapp.processEvents()
            time.sleep(0.1)
            self.qapp.processEvents()
            self.assertEqual(
                self._process_manager.get_dataset_state(
                    dataset_id=scan.get_dataset_identifier(),
                    process=self.widget,
                ),
                DatasetState.SUCCEED,
            )
